#ifndef ccbcontrolclient_h
#define ccbcontrolclient_h

#include "ccbclientlink.h"  /* CCBClientLink, CCBTelemetryStream, CCBConfig */

/*-----------------------------------------------------------------------*
 * This file defines control-link communications interface between       *
 * a CCB manager and a CCB server.                                       *
 *-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*
 *              The manager interface to the control link                *
 *-----------------------------------------------------------------------*
 *
 * CCBControlClient objects manage the client side of the control link
 * to a CCB server.
 */
typedef struct CCBControlClient CCBControlClient;

CCBControlClient *new_CCBControlClient(CCBClientLink *client);
CCBControlClient *del_CCBControlClient(CCBControlClient *cc);

/*
 * Queue a start-scan command to be sent to a the CCB server.
 */
int ccb_cc_queue_start_scan_cmd(CCBControlClient *cc, long id,
				const CCBConfig *cnf, unsigned long scan,
				unsigned long mjd, unsigned long tod);

/*
 * Queue a stop-scan command to be sent to the CCB server.
 */
int ccb_cc_queue_stop_scan_cmd(CCBControlClient *cc, long id,
			       const CCBConfig *cnf, unsigned long scan);

/*
 * Queue a dump-scan command to be sent to the CCB server.
 */
int ccb_cc_queue_dump_scan_cmd(CCBControlClient *cc, long id,
			       const CCBConfig *cnf, unsigned long scan,
			       unsigned short adc, unsigned long samples,
			       unsigned long frames);

/*
 * Queue a reset command to be sent to the CCB server.
 */
int ccb_cc_queue_reset_cmd(CCBControlClient *cc, long id);

/*
 * Queue an ping command to be sent to the CCB server.
 */
int ccb_cc_queue_ping_cmd(CCBControlClient *cc, long id, unsigned long seq);

/*
 * Queue a "load-driver" command to be sent to the CCB server.
 */
int ccb_cc_queue_load_driver_cmd(CCBControlClient *cc, long id,
				 CCBDriverType type);
/*
 * Queue a status-request command to be sent to the CCB server.
 */
int ccb_cc_queue_status_request_cmd(CCBControlClient *cc, long id);

/*
 * Queue a shutdown command to be sent to the CCB server.
 */
int ccb_cc_queue_shutdown_cmd(CCBControlClient *cc, long id);

/*
 * Queue a reboot command to be sent to the CCB server.
 */
int ccb_cc_queue_reboot_cmd(CCBControlClient *cc, long id);

/*
 * Queue a monitoring-control command to be sent to a given CCB server.
 */
int ccb_cc_queue_monitor_cmd(CCBControlClient *cc, long id,
			     unsigned short period);

/*
 * Queue a telemetry-stream control command to be sent to a given CCB server.
 */
int ccb_cc_queue_telemetry_cmd(CCBControlClient *cc, long id,
			       unsigned short streams);

/*
 * Queue a log-control command to be sent to a given CCB server.
 */
int ccb_cc_queue_logger_cmd(CCBControlClient *cc, long id,
			    unsigned long period);

/*
 * Register the callback function to be called whenever a given CCB client
 * object receives a control-status reply message.
 */
int ccb_cc_status_reply_callback(CCBControlClient *cc,
				      CCBStatusReplyFn *fn, void *data);

/*
 * Register the callback function that is used to report
 * control-command errors.
 */
int ccb_cc_cmd_error_callback(CCBControlClient *cc, CCBCmdErrorFn *fn,
			      void *data);
/*
 * Attempt to send all queued control messages to the CCB server.
 */
int ccb_cc_send_control(CCBControlClient *cc);

/*
 * Read up to one command message from the CCB client control socket,
 * and when a complete command-message has been read, deliver it to
 * the manager by calling the appropriate callback function.
 */
int ccb_cc_read_control(CCBControlClient *cc);

/*
 * Use the callback registered previously by calling
 * ccb_cc_cmd_error_callback(), to report a control-command error.
 */
int ccb_cc_command_error(CCBControlClient *cc, long id, CCBCmdStatus status);

/*
 * Initiate a connection to a CCB control server.
 */
int ccb_cc_connect(CCBControlClient *cc, const char *host, int nonblocking);

/*
 * Select between blocking and non-blocking control-client socket I/O.
 */
int ccb_cc_non_blocking_io(CCBControlClient *cc, int on);

/*
 * Return the file-descriptor of the telemetry socket.
 */
int ccb_cc_socket(CCBControlClient *cc);

/*
 * Disconnect from a remote CCB telemetry server.
 */
void ccb_cc_disconnect(CCBControlClient *cc);

/*
 * Enumerate the states that the control client can be in. Note that
 * due to the need to be alerted if the CCB server breaks the
 * connection, the client must always watch for the readability of the
 * control socket, so this requirement isn't enumerated.
 */
typedef enum {
  CCB_CONTROL_CLIENT_DOWN,/* Their is currently no control link */
  CCB_CONTROL_CLIENT_PEND,/* A non-blocking connect is in progress */
  CCB_CONTROL_CLIENT_IDLE,/* No command-messages are currently queued */
  CCB_CONTROL_CLIENT_BUSY /* The client is in the midst of sending a message */
} CCBControlClientStatus;

/*
 * Query the current status of the control client.
 */
CCBControlClientStatus ccb_cc_status(CCBControlClient *cc);

#endif
