/*
 * Standard C-library includes.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>

/*
 * Unix includes.
 */
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/*
 * Local includes.
 */
#include "ccbconstants.h"
#include "ccbserverlink.h"
#include "ccbdemodriver.h"

/*
 * Collect together the parameters that configure dump-mode scans.
 */
typedef struct {
  unsigned short adc;       /* The index of the ADC to be dumped */
  unsigned long samples;    /* The max number of samples per integration, */
                            /* or CCB_DUMP_MAX_SAMPLES for the hard limit. */
  unsigned long frames;     /* The number of per-integrations dump-frames */
                            /* to deliver, or CCB_DUMP_ALL_FRAMES to */
                            /*  dump frames indefinitely. */
} DemoDumpPars;

/*
 * Encapsulate the running parameters of the demo program. Note that a
 * typedef alias for this struct is defined in ccbdemodriver.c.
 */
struct CCBDemoDriverData {
  unsigned long ninteg;     /* The number of integrations that have been */
                            /*  started so far within the current scan. */
  unsigned long nmonitor;   /* The number of monitor messages sent so far */
  unsigned long scan;       /* The ID of the current scan */
  CCBConfig *cnf;           /* The CCB configuration of the current scan */
  CCBTimeStamp wakeup_time; /* The time of the next simulated event */
  int driver_loaded;        /* True when the device-driver is loaded */
  int scan_active;          /* True while a scan or intra-scan is running */
  int dump_mode;            /* True during a dump-mode intra-scan */
  DemoDumpPars dump;        /* The characteristics of a dump-mode scan */
  CCBInterval integ_dt;     /* The duration of integrations in the current */
                            /*  scan. */
  CCBTimeStamp integ_ts;    /* The time at which the most recent integration */
                            /*  period started. */
};

static CCB_DRIVER_LOAD_FN(demo_driver_load_fn);
static CCB_DRIVER_UNLOAD_FN(demo_driver_unload_fn);
static CCB_DRIVER_TELL_FN(demo_driver_tell_fn);
static CCB_REBOOT_RTC_FN(demo_reboot_rtc_fn);
static CCB_SHUTDOWN_RTC_FN(demo_shutdown_rtc_fn);
static CCB_DRIVER_SELECT_EVENTS_FN(demo_driver_select_events_fn);
static CCB_DRIVER_CHECK_EVENTS_FN(demo_driver_check_events_fn);

static int demo_queue_fake_integration(CCBServerLink *sl,
				       CCBDemoDriverData *dd);
static int demo_queue_fake_monitoring(CCBServerLink *sl,
				      CCBDemoDriverData *dd);
static int demo_display_config(CCBDemoDriverData *dd);
static int demo_prepare_scan(CCBDemoDriverData *dd, unsigned long scan,
			     unsigned long mjd, unsigned long tod,
			     DemoDumpPars *dump);
static int demo_config_scan(CCBDemoDriverData *dd, const CCBDrvConfScan *cmd);
static int demo_abort_scan(CCBDemoDriverData *dd);

/*.......................................................................
 * Create a new CCBDemoDriverData object. This contains all of the resources
 * needed by the demonstration CCB driver.
 *
 * Output:
 *  return  CCBDemoDriverData *  The new object, or NULL on error.
 */
CCBDemoDriverData *new_CCBDemoDriverData(void)
{
  CCBDemoDriverData *dd;  /* The object to be returned */
/*
 * Allocate the container.
 */
  dd = malloc(sizeof(CCBDemoDriverData));
  if(!dd) {
    fprintf(stderr, "new_CCBDemoDriverData: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBDemoDriverData().
 */
  dd->ninteg = 0;
  dd->nmonitor = 0;
  dd->scan = 0;
  dd->cnf = NULL;
  dd->wakeup_time.mjd = 0;
  dd->wakeup_time.sec = 0;
  dd->wakeup_time.ns = 0;
  dd->driver_loaded = 0;
  dd->scan_active = 0;
  dd->dump_mode = 0;
  ccb_zero_interval(&dd->integ_dt);
  ccb_get_timestamp(&dd->integ_ts);
/*
 * Allocate a CCB configuration object.
 */
  dd->cnf = new_CCBConfig();
  if(!dd->cnf)
    return del_CCBDemoDriverData(dd);
  return dd;
}

/*.......................................................................
 * Delete a CCBDemoDriverData object.
 *
 * Input:
 *  data     CCBDemoDriverData *  The object to be deleted.
 * Output:
 *  return  CCBDemoDriverData *  The deleted object (always NULL).
 */
CCBDemoDriverData *del_CCBDemoDriverData(CCBDemoDriverData *dd)
{
  if(dd) {
/*
 * Delete the CCB configuration object.
 */
    dd->cnf = del_CCBConfig(dd->cnf);
/*
 * Delete the container of the object.
 */
    free(dd);
  };
  return NULL;
}

/*.......................................................................
 * Allocate the interface object of the demo-driver.
 *
 * Input:
 *  data  CCBDemoDriverData *  The resource object of the driver.
 * Output:
 *  return  CCBServerDriver *  The new interface object, or NULL on error.
 */
CCBServerDriver *new_CCBDemoDriver(CCBDemoDriverData *dd)
{
  return new_CCBServerDriver(dd, demo_driver_load_fn,
			     demo_driver_unload_fn, demo_driver_tell_fn,
			     demo_driver_select_events_fn,
			     demo_driver_check_events_fn,
			     demo_reboot_rtc_fn, demo_shutdown_rtc_fn);
}

/*.......................................................................
 * Load the simulated CCB device driver.
 */
static CCB_DRIVER_LOAD_FN(demo_driver_load_fn)
{
  CCBDemoDriverData *dd = (CCBDemoDriverData *) data;
/*
 * This function isn't supposed to be called until any previously installed
 * driver has been unloaded.
 */
  if(dd->driver_loaded) {
    fprintf(stderr, "demo: Unexpected request to load the CCB driver.\n");
    ccb_record_server_faults(sl, CCB_SOFT_FAULT);
    return 1;
  };
/*
 * Mark the device driver as now loaded.
 */
  dd->driver_loaded = 1;
  printf("demo: CCB device-driver loaded.\n");
  return 0;
}

/*.......................................................................
 * Unload the simulated CCB device driver.
 */
static CCB_DRIVER_UNLOAD_FN(demo_driver_unload_fn)
{
  CCBDemoDriverData *dd = (CCBDemoDriverData *) data;
/*
 * Mark the device driver as now unloaded.
 */
  dd->driver_loaded = 0;
  printf("demo: CCB device-driver unloaded.\n");
  return 0;
}

/*.......................................................................
 * Execute a device-driver command.
 */
static CCB_DRIVER_TELL_FN(demo_driver_tell_fn)
{
  CCBDemoDriverData *dd = (CCBDemoDriverData *) data;
  DemoDumpPars dump;  /* Dump-mode configuration parameters */
/*
 * This function isn't supposed to be called when the device driver
 * isn't loaded.
 */
  if(!dd->driver_loaded) {
    fprintf(stderr,
	    "demo: Unexpected command while device driver not loaded.\n");
    ccb_record_server_faults(sl, CCB_SOFT_FAULT);
    return 1;
  };
/*
 * Determine what type of command has been sent, and execute it.
 */
  switch(cmd->type) {
  case CCB_DRV_ABORT_SCAN: /* Halt any ongoing scan or intra-scan */
    if(demo_abort_scan(dd))
      return 1;
    break;
  case CCB_DRV_CONF_SCAN:  /* Configure the next scan */
    if(demo_config_scan(dd, &cmd->pars.conf))
      return 1;
    break;
  case CCB_DRV_STAGE_SCAN: /* Start a new scan at a given time */
    if(demo_prepare_scan(dd, cmd->pars.stage.scan, cmd->pars.stage.mjd,
			 cmd->pars.stage.secs, NULL))
      return 1;
    break;
  case CCB_DRV_INTRA_SCAN: /* Start a new intra-scan ASAP */
    if(demo_prepare_scan(dd, cmd->pars.intra.scan, 0, 0, NULL))
      return 1;
    break;
  case CCB_DRV_DUMP_SCAN:  /* Start a new dump-mode intra-scan ASAP */
    dump.adc = cmd->pars.dump.adc;
    dump.samples = cmd->pars.dump.samples;
    dump.frames = cmd->pars.dump.frames;
    if(demo_prepare_scan(dd, cmd->pars.dump.scan, 0, 0, &dump))
      return 1;
    break;
  default:
    fprintf(stderr, "demo: Unknown driver command received.\n");
    return 1;
  };
  return 0;
}

/*.......................................................................
 * Stage a simulated reboot.
 */
static CCB_REBOOT_RTC_FN(demo_reboot_rtc_fn)
{
  printf("demo: Reboot requested.\n");
  return 1;
}

/*.......................................................................
 * Stage a simulated shutdown.
 */
static CCB_SHUTDOWN_RTC_FN(demo_shutdown_rtc_fn)
{
  printf("demo: Shutdown requested.\n");
  return 1;
}

/*.......................................................................
 * Print out the contents of a CCB configuration object.
 *
 * Input:
 *  data    CCBDemoDriverData *  The resources of the driver.
 * Output:
 *  return            int     0 - OK.
 *                            1 - Error.
 */
static int demo_display_config(CCBDemoDriverData *dd)
{
  CCBPhaseSwitchCnf phase;   /* The CCB phase-switch configuration */
  CCBCalDiodeCnf cal;        /* The CCB calibration diode configuration */
  CCBTimingCnf timing;       /* The CCB timing configuration */
  CCBSamplerCnf sampler;     /* The CCB sampler configuration */
  CCBInterval scan_dt;       /* An example scan duration */
  unsigned long ninteg;      /* The number of integrations that will fit into */
                             /*  the scan duration in scan_dt. */
  CCBInterval rem_dt;        /* Any time remainder after fitting ninteg */
                             /*  integrations into scan_dt. */
  CCBInterval integ_dt;      /* The integration time/bin/integration */
  int i;
/*
 * Look up the configuration parameters.
 */
  if(ccb_get_phase_switch_cnf(dd->cnf, &phase) ||
     ccb_get_cal_diode_cnf(dd->cnf, &cal) ||
     ccb_get_timing_cnf(dd->cnf, &timing) ||
     ccb_get_sampler_cnf(dd->cnf, &sampler))
    return 1;
/*
 * Display the phase-switch configuration.
 */
  printf("       Phase-switch configuration:\n"
         "        active_switches=%hu, closed_switches=%hu,\n"
         "        samp_per_state=%u\n",
	 phase.active_switches, phase.closed_switches, phase.samp_per_state);
/*
 * Display the cal-diode configuration.
 */
  printf("       Cal-diode configuration:\n"
	 "        ncal=%u,\n"
         "        diode_states={", cal.ncal);
  for(i=0; i<cal.ncal; i++)
    printf("%hu%s", cal.diode_states[i], i<cal.ncal-1 ? ", " : "");
  printf("}\n"
         "        diode_times={");
  for(i=0; i<cal.ncal; i++)
    printf("%lu%s", cal.diode_times[i], i<cal.ncal-1 ? ", " : "");
  printf("}\n");
/*
 * Display the timing configuration.
 */
  printf("       Timing configuration:\n"
         "        phase_switch_dt=%u, diode_rise_dt=%lu,\n"
         "        diode_fall_dt=%lu, integ_period=%lu,\n"
         "        roundtrip_dt=%u, holdoff_dt=%u, adc_delay_dt=%u\n",
	 timing.phase_switch_dt, timing.diode_rise_dt,
	 timing.diode_fall_dt, timing.integ_period,
         (unsigned) timing.roundtrip_dt, (unsigned) timing.holdoff_dt,
	 (unsigned) timing.adc_delay_dt);
/*
 * Display the sampler configuration.
 */
  printf("       Sampler configuration:\n"
         "        ");
  switch(sampler.sample_type) {
  case CCB_ADC_SAMPLES:
    printf("Real ADC");
    break;
  case CCB_FAKE_SAMPLES:
    printf("Fake pseudo-random");
    break;
  default:
    printf("(unknown! [%d])", (int)sampler.sample_type);
    break;
  };
  printf(" samples will be used.\n");
/*
 * Report how long each integration lasts.
 */
  printf("       The duration of each integration = %g seconds.\n",
     (double) dd->integ_dt.sec + (double) dd->integ_dt.ns / CCB_NS_PER_SEC);
/*
 * Report how many integrations will fit into a 1 minute scan.
 */
  scan_dt.sec = 60;
  scan_dt.ns = 0;
  if(ccb_integ_per_interval(dd->cnf, &scan_dt, &ninteg, &rem_dt))
    return 1;
  printf(
       "       A 1-minute scan would contain %lu integrations + %g seconds.\n",
	 ninteg, rem_dt.sec + (double)rem_dt.ns/CCB_NS_PER_SEC);
/*
 * Display the integration time per phase-switch bin, per integration.
 */
  ccb_integration_time(dd->cnf, &integ_dt);
  printf("      The integration time/phase-bin/integration: = %g seconds.\n",
	 integ_dt.sec + (double)integ_dt.ns/CCB_NS_PER_SEC);
  return 0;
}

/*.......................................................................
 * Queue a fake integration message to be sent to the manager.
 *
 * Input:
 *  sl        CCBServerLink *  The parent CCB server.
 *  data  CCBDemoDriverData *  The resources of the driver.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
static int demo_queue_fake_integration(CCBServerLink *sl,
				       CCBDemoDriverData *dd)
{
  unsigned long values[CCB_MAX_INTEG]; /* The fake integrations */
  CCBTimeStamp t;                      /* The current time */
  unsigned short flags = 0;            /* Integration status-flags */
  unsigned long integ_per_cycle;       /* Integrations per cal cycle */
  int i;
/*
 * Lookup the current time.
 */
  if(ccb_get_timestamp(&t))
    return 1;
/*
 * How many integrations are there per calibration cycle?
 */
  if(ccb_cal_cycle_length(dd->cnf, &integ_per_cycle))
    return 1;
/*
 * Determine whether the most recent cal-diode switch requires that we
 * flag the current integration.
 */
  if(integ_per_cycle > 0) {
    CCBCalDiodeCnf cal; /* The cal-diode configuration parameters */
    unsigned now;      /* The current cal-diode state-machine step */
    unsigned was;      /* The previous cal-diode state-machine step */
    CCBInterval dt;    /* The settling time of the cal diodes */
    CCBInterval rdt;   /* Fractional integration time at end of settling time */
    unsigned long integs_to_settle; /* The number of integrations to settle */
/*
 * Find out how far into a calibration cycle we have reached.
 */
    unsigned remaining_integs = (dd->ninteg-1) % integ_per_cycle;
/*
 * Get the configuration-diode parameters.
 */
    if(ccb_get_cal_diode_cnf(dd->cnf, &cal))
      return 1;
/*
 * Find which state-machine step the new integration resides in,
 * and how many integrations into that step it has reached.
 */
    for(now=0; now<cal.ncal && cal.diode_times[now]<=remaining_integs; now++)
      remaining_integs -= cal.diode_times[now];
/*
 * What was the previous state-machine step?
 */
    was = (now==0 ? cal.ncal : now) - 1;
/*
 * Compute the settling time corresponding to the change of state from
 * the previous to the current cal-diode steps.
 */
    if(ccb_settling_time(dd->cnf, cal.diode_states[was], cal.diode_states[now],
			 0U, 0U, &dt))
      return 1;
/*
 * Compute the rounded-up number of complete integrations corresponding to
 * the settling time.
 */
    if(ccb_integ_per_interval(dd->cnf, &dt, &integs_to_settle, &rdt))
      return 1;
/*
 * Round up the number of integrations.
 */
    if(!ccb_interval_is_zero(&rdt))
      integs_to_settle++;
/*
 * If enough time has passed since the start of the current cal-diode step,
 * such that the cal diodes have settled, flag the integration as usable.
 */
    if(remaining_integs >= integs_to_settle)
      flags |= CCB_INTEG_OK;
/*
 * Record the target states of the cal-diodes.
 */
    if(cal.diode_states[now] & CCB_CAL_DIODE_A)
      flags |= CCB_CAL_A_ON;
    if(cal.diode_states[now] & CCB_CAL_DIODE_B)
      flags |= CCB_CAL_B_ON;
  };
/*
 * Mark all of the slave FPGA boards as being present.
 */
  flags |= CCB_SLAVE0_OK | CCB_SLAVE1_OK | CCB_SLAVE2_OK | CCB_SLAVE3_OK;
/*
 * Fill the array of integrated values.
 */
  for(i=0; i<CCB_MAX_INTEG; i++)
    values[i] = UINT_MAX * (double)rand()/RAND_MAX;
/*
 * Queue the integration message.
 */
  if(ccb_queue_integ_msg(sl, &dd->integ_ts, dd->scan, dd->ninteg - 1,
			 flags, values, CCB_MAX_INTEG))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a fake monitoring message to be sent to the manager.
 *
 * Input:
 *  data    CCBDemoDriverData *  The resources of the driver.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int demo_queue_fake_monitoring(CCBServerLink *sl,
				      CCBDemoDriverData *dd)
{
  unsigned long values[CCB_MAX_MONITOR]; /* The fake monitoring values */
  CCBTimeStamp t;                        /* The current time */
  int i;
/*
 * Lookup the current time.
 */
  if(ccb_get_timestamp(&t))
    return 1;
/*
 * Fill the array of monitoring values.
 */
  for(i=0; i<CCB_MAX_MONITOR; i++)
    values[i] = UINT_MAX * (double)rand()/RAND_MAX;
/*
 * Queue the monitor-data message.
 */
  if(ccb_queue_monitor_msg(sl, &t, dd->scan, dd->nmonitor++,
			   values, CCB_MAX_MONITOR))
    return 1;
  return 0;
}

/*.......................................................................
 * Prepare the driver to start a new scan or intra-scan.
 *
 * Input:
 *  data    CCBDemoDriverData *  The resources of the driver.
 *  scan   unsigned long    The numeric identifyier of the new scan.
 *  mjd    unsigned long    If none zero, this specifies the date
 *                          at which to start the scan, expressed as
 *                          a Modified Julian Day number.
 *  tod    unsigned long    If mjd!=0, this is the time of day at which
 *                          to start the scan, specified as the number
 *                          of seconds into the day.
 *  dump    DemoDumpPars *  If not NULL, then the scan is to be a dump-mode
 *                          scan, and the parameters contained in this
 *                          object denote the configuration of the dump.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int demo_prepare_scan(CCBDemoDriverData *dd, unsigned long scan,
			     unsigned long mjd, unsigned long tod,
			     DemoDumpPars *dump)

{
  CCBInterval dt;   /* The duration of the delay preceding the integration */
/*
 * Get the name by which to refer to the requested type of scan.
 */
  const char *name = mjd==0 ? (dump ? "dump-mode intra-scan" : "intra-scan") :
    "scan";
/*
 * Start-scan and intra-scan commands aren't supposed to be invoked
 * when the driver is unloaded.
 */
  if(!dd->driver_loaded) {
    fprintf(stderr, "demo: New %s requested while driver not loaded.\n", name);
    return 1;
  };
/*
 * Start-scan and intra-scan commands aren't supposed to be invoked
 * until any previous scan or intra-scan has been aborted.
 */
  if(dd->scan_active) {
    fprintf(stderr,"demo: Not starting %s, because a scan is already active.\n",
	    name);
    return 1;
  };
/*
 * Report what we are doing.
 */
  printf("demo: Preparing to start %s %lu.\n", name, scan);
/*
 * Record the scan ID counter and reset the integration and monitor counters.
 */
  dd->scan = scan;
  dd->ninteg = 0;
  dd->nmonitor = 0;
  dd->scan_active = 1;
  if(dump) {
    dd->dump_mode = 1;
    dd->dump = *dump;
  };
/*
 * Do we need to wait before starting the scan?
 */
  if(mjd != 0) {
    dd->wakeup_time.mjd = mjd;
    dd->wakeup_time.sec = tod;
    dd->wakeup_time.ns = 0;
  } else {  /* Specify the current time as the next select() wakeup time */
    if(ccb_get_timestamp(&dd->wakeup_time))
      return 1;
  };
/*
 * Compute how long we will need to wait before starting the new scan.
 */
  if(ccb_time_until(dd->wakeup_time, &dt))
    return 1;
/*
 * Report how long we have to wait before starting the new scan.
 */
  printf("demo: Waiting %g seconds until start-time of %s.\n",
	 (double) dt.sec + (double) dt.ns / CCB_NS_PER_SEC, name);
  return 0;
}

/*.......................................................................
 * Reconfigure the simulated device driver.
 *
 * Input:
 *  data         CCBDemoDriverData *  The resources of the driver.
 *  cmd  const CCBDrvConfScan *  The new configuration.
 * Output:
 *  return                int    0 - OK.
 *                               1 - Error.
 */ 
static int demo_config_scan(CCBDemoDriverData *dd, const CCBDrvConfScan *cmd)
{
/*
 * Configuration commands aren't supposed to be invoked when the
 * driver is unloaded.
 */
  if(!dd->driver_loaded) {
    fprintf(stderr,
	    "demo: Scan reconfiguration requested while driver not loaded.\n");
    return 1;
  };
/*
 * This function is only suposed to be called after the most recent scan
 * has been aborted.
 */
  if(dd->scan_active) {
    fprintf(stderr, "demo: Rejecting re-configuration during scan.\n");
    return 1;
  };
/*
 * Copy the configuration into a CCB configuration object.
 */
  printf("demo: Installing new scan configuration:\n");
  if(ccb_set_config(dd->cnf, &cmd->phase, &cmd->cal, &cmd->timing,
		    &cmd->sampler)) {
    fprintf(stderr, "demo: Configuration error.\n");
    return 1;
  };
/*
 * Compute the duration of each integration in the new scan.
 */
  if(ccb_integration_duration(dd->cnf, &dd->integ_dt))
    return 1;
/*
 * Report the new configuration.
 */
  if(demo_display_config(dd))
    return 1;
  return 0;
}

/*.......................................................................
 * Abort the current scan or intra-scan, if any.
 *
 * Input:
 *  data         CCBDemoDriverData *  The resources of the driver.
 * Output:
 *  return                int    0 - OK.
 *                               1 - Error.
 */ 
static int demo_abort_scan(CCBDemoDriverData *dd)
{
/*
 * Abort-scan commands aren't supposed to be invoked when the
 * driver isn't loaded.
 */
  if(!dd->driver_loaded) {
    fprintf(stderr, "demo: Scan abort requested while driver not loaded.\n");
    return 1;
  };
/*
 * Mark the current scan/intra-scan as stopped.
 */
  dd->scan_active = 0;
  return 0;
}

/*.......................................................................
 * Update the specified select() arguments to register the demo inactivity
 * timeout. This timeout is used to provide the simulator clock.
 */
static CCB_DRIVER_SELECT_EVENTS_FN(demo_driver_select_events_fn)
{
  CCBDemoDriverData *dd = (CCBDemoDriverData *) data;
  CCBInterval dt;         /* The timeout duration */
/*
 * Compute the amount of time remaining until the next wakeup time.
 */
  if(ccb_time_until(dd->wakeup_time, &dt))
    return 1;
/*
 * Record the new select() timeout if there are any integration or
 * monitoring events to simulate.
 */
  if(dd->scan_active) {
    timeout->tv_sec = dt.sec;
    timeout->tv_usec = dt.ns / CCB_NS_PER_USEC;
  };
  return 0;
}

/*.......................................................................
 * Respond to a select() inactivity timeout.
 */
static CCB_DRIVER_CHECK_EVENTS_FN(demo_driver_check_events_fn)
{
  CCBDemoDriverData *dd = (CCBDemoDriverData *) data;
/*
 * Did select() timeout?
 */
  if(!rfds && !wfds) {
/*
 * Is this the start of the first integration period of a new scan?
 */
    if(dd->ninteg==0) {
/*
 * Report the characteristics of a new dump-mode scan,
 * at the start of the first integration.
 */
      if(dd->dump_mode) {
	printf("demo: Starting dump of ");
	if(dd->dump.samples == CCB_DUMP_MAX_SAMPLES)
	  printf("maximal");
	else
	  printf("up to %lu", dd->dump.samples);
	printf(" samples/integration of ADC %u, ", dd->dump.adc);
	if(dd->dump.frames == CCB_DUMP_ALL_FRAMES)
	  printf("indefinitely.\n");
	else
	  printf("for %lu integrations\n", dd->dump.frames);
      } else {
	printf("demo: Starting scan.\n");
      };
/*
 * Make the time at which the first integration was started be the
 * base time wrt which future wakeup times are computed.
 */
      if(ccb_get_timestamp(&dd->wakeup_time))
	return 1;
/*
 * Have we reached the end of an integration period?
 */
    } else {
/*
 * Queue a fake integration-data message to be sent to the manager.
 * Don't do this while in dump mode.
 */
      if(!dd->dump_mode && demo_queue_fake_integration(sl, dd))
	return 1;
/*
 * Queue a fake monitoring packet.
 */
      if(demo_queue_fake_monitoring(sl, dd))
	return 1;
    };
/*
 * Set the time at which select() should next timeout, to the current
 * time plus the duration of the next integration.
 */
    if(ccb_add_to_timestamp(&dd->wakeup_time, &dd->integ_dt,
			    &dd->wakeup_time))
      return 1;
/*
 * Increment the count of the number of integrations that have been
 * started within the current scan.
 */
    dd->ninteg++;
/*
 * Record the start time of the integration period that is now starting.
 */
    if(ccb_get_timestamp(&dd->integ_ts))
      return 1;
/*
 * In dump-mode, report when we reach the end of dumping the requested
 * number of integration's worth of samples.
 */
    if(dd->dump_mode && dd->dump.frames == dd->ninteg)
      printf("demo: Finished dumping samples from %lu integration periods.\n",
	     dd->dump.frames);
  };
  return 0;
}
