#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "cksum.h"
#include "ccblogtrail.h"
#include "ccbserverlink.h"     /* CCB_LOGID() */

/*
 * The checksums of a limited number of messages generated by a particular
 * logging statement are recorded in a structures of the following type.
 */
typedef struct {
  int nmsg;            /* The number of messages whose checksums are recorded */
  unsigned long *msgs; /* The array of 'maxmsg' checksums */
} CCBStmtMsgs;

/*
 * CCBLogTrail objects keep a record of the checksums of up to a given
 * number of message strings per logging statement, and thus allow the
 * caller to check for redundantly or excessively reported messages
 * from a given statement.
 */
struct CCBLogTrail {
  CheckSum *cs;            /* The checksum computation object */
  CCBStmtMsgs *stmt;       /* The array of 'maxstmt' per-statement message */
                           /*  lists. */
  size_t maxstmt;          /* The number of logging statements */
  size_t maxmsg;           /* The number of messages allowed per statement */
  CCBTimeStamp purge_time; /* The next time at which the checksums should be */
                           /*  purged. */
  CCBInterval purge_dt;    /* The time interval at which to purge the */
                           /*  checksums. */
};

static int ccb_lt_schedule_purge(CCBLogTrail *lt);

/*.......................................................................
 * Create a new CCBLogTrail object.
 *
 * Input:
 *  maxstmt      size_t    The maximum number of logging statements to
 *                         accomodate.
 *  maxmsg       size_t    The maximum number of messages allowed per
 *                         statement.
 * Output:
 *  return  CCBLogTrail *  The new object, or NULL on error.
 */
CCBLogTrail *new_CCBLogTrail(size_t maxstmt, size_t maxmsg)
{
  CCBLogTrail *lt;  /* The object to be returned */
  int i;
/*
 * Check the arguments.
 */
  if(maxstmt < 1 || maxmsg < 1) {
    errno = EINVAL;
    return NULL;
  };
/*
 * Allocate the container.
 */
  lt = malloc(sizeof(CCBLogTrail));
  if(!lt) {
    errno = ENOMEM;
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBLogTrail().
 */
  lt->cs = NULL;
  lt->stmt = NULL;
  lt->maxstmt = maxstmt;
  lt->maxmsg = maxmsg;
  lt->purge_time.mjd = 0;
  lt->purge_time.sec = 0;
  lt->purge_dt.sec = CCB_LOG_PURGE_DT;
  lt->purge_dt.ns = 0;
/*
 * Allocate the checksum computation object.
 */
  lt->cs = new_CheckSum();
  if(!lt->cs)
    return del_CCBLogTrail(lt);
/*
 * Allocate the array of statement message lists.
 */
  lt->stmt = (CCBStmtMsgs *) malloc(sizeof(*lt->stmt) * maxstmt);
  if(!lt->stmt) {
    errno = ENOMEM;
    return del_CCBLogTrail(lt);
  };
/*
 * Allocate one big array of checksums to be partioned into the
 * smaller per-statement arrays.
 */
  lt->stmt[0].msgs = (unsigned long *) malloc(sizeof(unsigned long) *
					      maxstmt * maxmsg);
  if(!lt->stmt[0].msgs) {
    errno = ENOMEM;
    return del_CCBLogTrail(lt);
  };
/*
 * Share the big array of checksums between the statements.
 */
  for(i=1; i<maxstmt; i++)
    lt->stmt[i].msgs = lt->stmt[i-1].msgs + maxmsg;
/*
 * Initialize the checksum lists to be empty, then schedule the first
 * purge of the table.
 */
  if(ccb_lt_purge(lt))
    return del_CCBLogTrail(lt);
  return lt;
}

/*.......................................................................
 * Delete a CCBLogTrail object.
 *
 * Input:
 *  lt     CCBLogTrail *  The object to be deleted.
 * Output:
 *  return CCBLogTrail *  The deleted object (always NULL).
 */
CCBLogTrail *del_CCBLogTrail(CCBLogTrail *lt)
{
  if(lt) {
/*
 * Delete the check-sum computation object.
 */
    lt->cs = del_CheckSum(lt->cs);
/*
 * Delete the array of statements.
 */
    if(lt->stmt) {
      if(lt->stmt[0].msgs)      /* Delete the combined arrays of checksums */
	free(lt->stmt[0].msgs);
      free(lt->stmt);
    };
/*
 * Finally, delete the container.
 */
    free(lt);
  };
  return NULL;
}

/*.......................................................................
 * Discard all recorded message checksums.
 *
 * Input:
 *  lt    CCBLogTrail *  The table of previously seen messages.
 * Output:
 *  return        int    0 - OK.
 *                       1 - Error.
 */
int ccb_lt_purge(CCBLogTrail *lt)
{
  int i;
  if(lt) {
    for(i=0; i<lt->maxstmt; i++)
      lt->stmt[i].nmsg = 0;
  };
/*
 * Schedule the next purge.
 */
  if(ccb_lt_schedule_purge(lt))
    return 1;
  return 0;
}

/*.......................................................................
 * Compute the checksum of a new message generated by a given
 * statement, and compare it to the checksums of previously registered
 * messages from that statement. If the message has been reported
 * before, or too many messages have been reported by the specified
 * statement, return 0 to indicate that the message should be
 * discarded. Otherwise return non-zero.
 *
 * Input:
 *  lt           CCBLogTrail *  The table of previously seen messages.
 *  statement  unsigned long    The unique id of the originating statement.
 *  text          const char *  The text of the message being reported by
 *                              the statement.
 * Output:
 *  return               int    0 - Discard the message.
 *                              1 - Report the message.
 */
int ccb_lt_verify_msg(CCBLogTrail *lt, unsigned long statement,
		      const char *text)
{
  CCBStmtMsgs *stmt;  /* The list of previously reported messages */
  unsigned long sum;  /* The checksum of text[] */
  int i;
/*
 * Check the arguments.
 */
  if(!lt || statement >= lt->maxstmt || !text) {
/*
 * Statement ID out of range?
 */
    if(statement >= lt->maxstmt) {
      const char *errmsg = "ccb_lt_verify_msg: Log-statement ID out of range.";
/*
 * If our statement ID is also bad, report to stderr. Otherwise, arrange
 * to log the error, by replacing the erroneously ID'd message with our own.
 */
      unsigned long id = CCB_LOGID(27);
      if(id >= lt->maxstmt) {
	fprintf(stderr, "%s\n", errmsg);
	errno = EINVAL;
	return 0;
      } else {
	statement = id;
	text = errmsg;
      };
    };
  };
/*
 * Is periodic purging enabled?
 */
  if(lt->purge_dt.sec > 0) {
    CCBInterval dt;     /* The remaining time until the scheduled purge */
/*
 * How long remains before the scheduled time of the next purge is
 * reached?
 */
    if(ccb_time_until(lt->purge_time, &dt))
      return 0;
/*
 * If we have reached or passed the scheduled purge time, purge the
 * recorded checksums.
 */
    if(dt.sec==0 && dt.ns==0 && ccb_lt_purge(lt))
      return 1;
  };
/*
 * Get the message recording node of the given statement.
 */
  stmt = lt->stmt + statement;
/*
 * If an excessive number of messages have already been reported by
 * the specified statement, reject the new one.
 */
  if(stmt->nmsg+1 >= lt->maxmsg)
    return 0;
/*
 * Compute the checksum of the message.
 */
  sum = checksum_of_object(lt->cs, text, strlen(text));
/*
 * Search the array of the checksums of already recorded messages to see
 * if the new message has already been reported. If it has, tell the
 * caller to discard it.
 */
  for(i=0; i<stmt->nmsg; i++) {
    if(sum == stmt->msgs[i])
      return 0;
  };
/*
 * The message appears not to have been reported by this statement since
 * the table was last purged, so record its checksum and tell the caller
 * to report the message.
 */
  stmt->msgs[stmt->nmsg++] = sum;
  return 1;
}

/*.......................................................................
 * Compute the time after which the checksums should be purged before
 * another log message is checked.
 *
 * Input:
 *  lt           CCBLogTrail *  The table of previously seen messages.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
static int ccb_lt_schedule_purge(CCBLogTrail *lt)
{
/*
 * Get the current time.
 */
  if(ccb_get_timestamp(&lt->purge_time))
    return 1;
/*
 * Add the current purge interval to it.
 */
  if(ccb_add_to_timestamp(&lt->purge_time, &lt->purge_dt, &lt->purge_time))
    return 1;
  return 0;
}

/*.......................................................................
 * Specify the interval at which recorded checksums should be purged.
 *
 * Input:
 *  lt           CCBLogTrail *  The table of previously seen messages.
 *  seconds    unsigned long    The interval at which old log messages
 *                              are forgotten, or 0 to turn off automatic
 *                              purging.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
int ccb_lt_lifetime(CCBLogTrail *lt, unsigned long seconds)
{
/*
 * Check the arguments.
 */
  if(!lt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Record the new interval.
 */
  lt->purge_dt.sec = seconds;
  lt->purge_dt.ns = 0;
/*
 * Schedule the next purge.
 */
  if(ccb_lt_schedule_purge(lt))
    return 1;
  return 0;
}

