#include <string.h>
#include <errno.h>

#include "ccbnetobj.h"
#include "ccbprint.h"

/*.......................................................................
 * Pack a network object into a network buffer for transmission.
 *
 * Input:
 *  types  CCBNetMsgTable *   A table of object definitions.
 *  net            NetBuf *   The network buffer in which to pack the object.
 *                            Note that the caller is responsible for calling
 *                            net_start_put() and net_end_put().
 *  msg         CCBNetMsg *   A pointer to the initial CCBNetMsg member
 *                            of the message. The msg->type member must
 *                            be the index of the message type within the
 *                            types[] array.
 * Output:    
 *  return            int     0 - OK.
 *                            1 - Error.
 */
int ccb_obj_to_net(const CCBNetMsgTable *types, NetBuf *net, CCBNetMsg *msg)
{
  const CCBNetMsgInfo *info;   /* The definition of the object being packed */
  int i;
/*
 * Check arguments.
 */
  if(!types || !net || !msg) {
    errno = EINVAL;
    return 1;
  };
/*
 * Look up the object definition.
 */
  info = ccb_net_obj_info(types, msg->type);
  if(!info)
    return 1;
/*
 * Write the header of the message.
 */
  if(net_start_put(net, msg->type))
    return 1;
/*
 * Pack the members of the network object sequentially.
 */
  for(i=1; i<info->nmember; i++) {
    const CCBNetMsgMember *member = info->member + i;
/*
 * Get a pointer to the start of the member in the object.
 */
    void *data = (char *)msg + member->offset;
/*
 * Write the member according to its type.
 */
    switch(member->type) {
    case CCB_NET_ASCII:
      {
	long len = strlen(data);
	if(net_put_long(net, 1, &len) ||
	   net_put_char(net, len, data))
	  return 1;
      };
      break;
    case CCB_NET_BYTE:
    case CCB_NET_UBYTE:
      if(net_put_char(net, member->ntype, data))
	return 1;
      break;
    case CCB_NET_SHORT:
    case CCB_NET_USHORT:
      if(net_put_short(net, member->ntype, data))
	return 1;
      break;
    case CCB_NET_LONG:
    case CCB_NET_ULONG:
      if(net_put_long(net, member->ntype, data))
	return 1;
      break;
    case CCB_NET_FLOAT:
      if(net_put_float(net, member->ntype, data))
	return 1;
      break;
    case CCB_NET_DOUBLE:
      if(net_put_double(net, member->ntype, data))
	return 1;
      break;
    default:
      errno = EINVAL;
      return 1;
      break;
    };
  };
/*
 * Finalize the message.
 */
  if(net_end_put(net))
    return 1;
  return 0;
}

/*.......................................................................
 * Un-pack a network object from a network buffer.
 *
 * Input:
 *  types  CCBNetMsgTable *   A table of object definitions.
 *  net            NetBuf *   The network buffer containing the packed object.
 *                            Note that the caller is responsible for calling
 *                            net_start_get() and net_end_get().
 *  id                int     The index of the object definition in 'types'.
 * Input/Output:
 *  msg         CCBNetMsg *   A pointer to the initial CCBNetMsg member
 *                            of a union of all possible message types.
 * Output:
 *  return            int     0 - OK.
 *                            1 - Error.
 */
int ccb_net_to_obj(const CCBNetMsgTable *types, NetBuf *net, CCBNetMsg *msg)
{
  const CCBNetMsgInfo *info; /* The definition of the object being un-packed */
  int type;                  /* The type of message to decode */
  int i;
/*
 * Check arguments.
 */
  if(!types || !net || !msg) {
    errno = EINVAL;
    return 1;
  };
/*
 * Get the type of the message from the header.
 */
  if(net_start_get(net, &type))
    return 1;
/*
 * Look up the object definition.
 */
  info = ccb_net_obj_info(types, type);
  if(!info)
    return 1;
/*
 * Record the message type in the message header.
 */
  msg->type = type;
/*
 * Un-pack the members of the network object sequentially.
 */
  for(i=1; i<info->nmember; i++) {
    const CCBNetMsgMember *member = info->member + i;
/*
 * Get a pointer to the start of the member in the object.
 */
    void *data = (char *)msg + member->offset;
/*
 * Extract the member according to its type.
 */
    switch(member->type) {
    case CCB_NET_ASCII:
      {
	long len;
	if(net_get_long(net, 1, &len))
	  return 1;
	if(len >= member->ntype) {
	  errno = EMSGSIZE;
	  return 1;
	};
	if(net_get_char(net, len, data))
	  return 1;
	((char *) data)[len] = '\0';
      };
      break;
    case CCB_NET_BYTE:
    case CCB_NET_UBYTE:
      if(net_get_char(net, member->ntype, data))
	return 1;
      break;
    case CCB_NET_SHORT:
    case CCB_NET_USHORT:
      if(net_get_short(net, member->ntype, data))
	return 1;
      break;
    case CCB_NET_LONG:
    case CCB_NET_ULONG:
      if(net_get_long(net, member->ntype, data))
	return 1;
      break;
    case CCB_NET_FLOAT:
      if(net_get_float(net, member->ntype, data))
	return 1;
      break;
    case CCB_NET_DOUBLE:
      if(net_get_double(net, member->ntype, data))
	return 1;
      break;
    default:
      errno = EINVAL;
      return 1;
      break;
    };
  };
/*
 * Check for unexpected trailing bytes.
 */
  if(net_end_get(net))
    return 1;
  return 0;
}

/*.......................................................................
 * Return the amount of buffer space needed to pack the largest of the
 * objects from the given table of object definitions.
 *
 * Input:
 *  types   CCBNetMsgTable *  The table of object definitions.
 * Output:
 *  return          size_t    The required network buffer space (bytes), or
 *                            0 on error.
 */
size_t ccb_net_max_obj_size(const CCBNetMsgTable *types)
{
  size_t max_size = 0;   /* The max object size */
  int i;
  for(i=0; i<types->nmsg; i++) {
    size_t size = ccb_net_obj_size(types, i);
    if(size > max_size)
      max_size = size;
  };
  return max_size;
}

/*.......................................................................
 * Return the packed size of a given network object.
 *
 * Input:
 *  types    CCBNetMsgTable *  The table of object descriptions.
 *  id                  int    The index of the object in the table.
 * Output:
 *  return           size_t    The packed size of the object. This will
 *                             be zero if the object contains no members,
 *                             or on error.
 */
size_t ccb_net_obj_size(const CCBNetMsgTable *types, int id)
{
  size_t size = 0;             /* The packed size of the object */
  const CCBNetMsgInfo *info;   /* The object definition */
  int i;
/*
 * Lookup the object definition.
 */
  info = ccb_net_obj_info(types, id);
  if(!info)
    return 0;
/*
 * Start with the size of the header that is prepended to all
 * messages.
 */
  size = NET_PREFIX_LEN;
/*
 * Add up the network sizes of the members of the object.
 */
  for(i=1; i<info->nmember; i++) {
    const CCBNetMsgMember *member = info->member + i;
    size += ccb_net_type_size(member->type) * member->ntype;
/*
 * Account for the extra length field that prefixes ASCII strings.
 */
    if(member->type == CCB_NET_ASCII)
      size += ccb_net_type_size(CCB_NET_LONG);
  };
  return size;
}

/*.......................................................................
 * Return the packed size of a given network data-type, in bytes.
 *
 * Input:
 *  type  CCBNetDataType   The type who's size is to be returned.
 * Output:
 *  return        size_t   The packed size of the data-type in bytes,
 *                         or 0 if the type is not known.
 */
size_t ccb_net_type_size(CCBNetDataType type)
{
  switch(type) {
  case CCB_NET_ASCII:
  case CCB_NET_BYTE:
  case CCB_NET_UBYTE:
    return NET_CHAR_SIZE;
  case CCB_NET_SHORT:
  case CCB_NET_USHORT:
    return NET_SHORT_SIZE;
  case CCB_NET_LONG:
  case CCB_NET_ULONG:
    return NET_LONG_SIZE;
  case CCB_NET_FLOAT:
    return NET_FLOAT_SIZE;
  case CCB_NET_DOUBLE:
    return NET_DOUBLE_SIZE;
  case CCB_NET_MSG:
    return NET_LONG_SIZE;
  };
/*
 * Unknown data-type.
 */
  errno = EINVAL;
  return 0;
}

/*.......................................................................
 * Return the native size of a given network object on the local
 * architecture.
 *
 * Input:
 *  types    CCBNetMsgTable *  The table of object descriptions.
 *  id                  int    The index of the object in the table.
 * Output:
 *  return           size_t    The native size of the object. This will
 *                             be zero if the object doesn't contain any
 *                             members (or if the object is unknown).
 */
size_t ccb_net_native_size(const CCBNetMsgTable *types, int id)
{
  const CCBNetMsgInfo *info;      /* The object definition */
/*
 * Lookup the object definition.
 */
  info = ccb_net_obj_info(types, id);
  if(!info)
    return 0;
  return info->native_size;
}

/*.......................................................................
 * Look up the definition of a given network object type.
 *
 * Input:
 *  types   CCBNetMsgTable *  A table of object definitions.
 *  id                 int    The type of object, as an index into
 *                            types->objs[].
 * Output:
 *  return   CCBNetMsgInfo *  The definition of the object, or NULL if
 *                            types==NULL or id < 0 or id >= types->nmsg.
 */
const CCBNetMsgInfo *ccb_net_obj_info(const CCBNetMsgTable *types, int id)
{
  if(!types || id < 0 || id >= types->nmsg) {
    errno = EINVAL;
    return NULL;
  };
  return types->info + id;
}

/*.......................................................................
 * Return the name of a given type of network object.
 *
 * Input:
 *  types    CCBNetMsgTable *  The table of object descriptions.
 *  id                  int    The index of the object in the table.
 * Output:
 *  return       const char *  The name of the object, or the string
 *                             "unknown".
 */
const char *ccb_net_obj_name(const CCBNetMsgTable *types, int id)
{
  const CCBNetMsgInfo *info;   /* The object definition */
/*
 * Lookup the object definition.
 */
  info = ccb_net_obj_info(types, id);
  return info ? info->name : "unknown";
}

/*.......................................................................
 * Check the contents of a table of message descriptions.
 *
 * Input:
 *  types const CCBNetMsgTable *  The table of message-structure
 *                                descriptions to be checked.
 *  errbuf                char *  A buffer in which to write error
 *                                messages regarding the contents of
 *                                the table.
 *  errdim              size_t    The allocated size of errbuf[].
 * Output:
 *  return                 int    0 - OK.
 *                                1 - Error (see the error message
 *                                    in the errbuf[] argument).
 */
int ccb_check_net_table(const CCBNetMsgTable *types, char *errbuf,
			size_t errdim)
{
  int i,j;
/*
 * Check the arguments.
 */
  if(!types) {
    ccb_snprintf(errbuf, errdim, "ccb_check_net_table: NULL argument(s).\n");
    return 1;
  };
/*
 * Check the top level table fields.
 */
  if(!types->name) {
    ccb_snprintf(errbuf, errdim, "Missing table name");
    return 1;
  };
  if(!types->info) {
    ccb_snprintf(errbuf, errdim, "Missing array of message descriptions");
    return 1;
  };
  if(types->union_size < 1 && types->nmsg>0) {
    ccb_snprintf(errbuf, errdim, "Zero message-structure union size");
    return 1;
  };
/*
 * Check the individual message descriptions.
 */
  for(i=0; i<types->nmsg; i++) {
    const CCBNetMsgInfo *info = types->info + i;
/*
 * Check that the message has a name, so that we can use the name in
 * error reports below.
 */
    if(!info->name) {
      ccb_snprintf(errbuf, errdim, "Message %d has no name", i+1);
      return 1;
    };
/*
 * Check the top-level description of the message.
 */
    if(!info->member || info->nmember<1) {
      ccb_snprintf(errbuf, errdim, "Message \"%s\" has no members", info->name);
      return 1;
    };
    if(!info->native_size) {
      ccb_snprintf(errbuf, errdim, "Illegal zero native size of message \"%s\"",
		   info->name);
      return 1;
    };
/*
 * Check each of the members of the message.
 */
    for(j=0; j<info->nmember; j++) {
      const CCBNetMsgMember *m = info->member + j;
/*
 * Check that the member has a name (we need this for error reporting below).
 */
      if(!m->name) {
	ccb_snprintf(errbuf, errdim, "Member %d of message \"%s\" has no name",
		     j+1, info->name);
	return 1;
      };
/*
 * Check the datatype of the member.
 */
      switch(m->type) {
      case CCB_NET_MSG:    /* The required initial CCBNetMsg header */
	if(j != 0 || m->offset != 0) {
	  ccb_snprintf(errbuf, errdim,
		       "Misplaced CCBNetMsg header in \"%s.%s\"",
		       info->name, m->name);
	  return 1;
	};
	break;
      case CCB_NET_ASCII:  /* Elements of a C string */
      case CCB_NET_BYTE:   /* 8-bit signed integers */
      case CCB_NET_UBYTE:  /* 8-bit unsigned integers */
      case CCB_NET_SHORT:  /* 16-bit signed integers */
      case CCB_NET_USHORT: /* 16-bit unsigned integers */
      case CCB_NET_LONG:   /* 32-bit signed integers */
      case CCB_NET_ULONG:  /* 32-bit unsigned integers */
      case CCB_NET_FLOAT:  /* 32-bit floating point */
      case CCB_NET_DOUBLE: /* 64-bit floating point */
	if(j == 0) {
	  ccb_snprintf(errbuf, errdim,
		       "Missing CCBNetMsg header in message \"%s\"",
		       info->name);
	  return 1;
	};
	break;
      default:
	ccb_snprintf(errbuf, errdim, "Member \"%s.%s\" has an unknown type",
		     info->name, m->name);
	return 1;
	break;
      };
/*
 * All members must have at least one element.
 */
      if(m->ntype < 1) {
	ccb_snprintf(errbuf, errdim, "Zero number of elements in \"%s.%s\"",
		     info->name, m->name);
	return 1;
      };
/*
 * The header member must have precisely one element.
 */
      if(j==0 && m->ntype != 1) {
	ccb_snprintf(errbuf, errdim,
		     "Dimension of header of \"%s\" should be 1.",
		     info->name);
	return 1;
      };
    };
  };
/*
 * The table appears to be sane.
 */
  return 0;
}
