#ifndef ccbprint_h
#define ccbprint_h

#include <stdarg.h>

/*
 * This module provides a facility that parses printf-style formats
 * and sequentially passes each converted argument and intervening
 * string segments to a given caller-provided callback function. The
 * caller can also provide anonymous data to be passed to their output
 * function via a (void *context) argument.
 */

/*
 * The caller-provided output function should be prototyped and declared
 * with the following macro. The arguments of the function are:
 *
 * Input:
 *  buffer           char *   The rendered output of the latest converted
 *                            argument or intervening string segment. This
 *                            buffer is NOT '\0' terminated.
 *  nc                int     The length of the string prefix to take from
 *                            the buffer argument.
 *  context          void *   The context argument of ccb_print_format().
 * Output:
 *  return            int     0 - OK.
 *                            1 - An error occured. This tells
 *                                ccb_print_format() to abort and return -1.
 */
#define CCB_PRINT_FN(fn) int (fn)(char *buffer, int nc, void *context)

/*.......................................................................
 * Parse a printf-style format string and send the formatted output to a
 * caller provided callback function.
 *
 * Input:
 *  fmt               char *  The printf() format to use to convert the
 *                            arguments in 'ap'.
 *  ap             va_list    The list of arguments to associate with fmt[].
 *                            It is the responsibility of the caller to
 *                            call va_start() before calling print_format(),
 *                            and va_end() after print_format() returns.
 *  print_fn  CCB_PRINT_FN(*) The function to pass converted output to.
 *  context           void *  A pointer to any data that the caller wishes
 *                            to have access to within print_fn().
 * Output:
 *  return             int    The number of bytes passed to print_fn(), or
 *                            -1 on error.
 */ 
int ccb_print_format(const char *fmt, va_list ap, CCB_PRINT_FN(*print_fn),
		     void *context);

/*.......................................................................
 * Provide a portable equivalent to the sparsely available snprintf()
 * function. This function differs from sprintf() in that it takes an
 * aditional output buffer length so that buffer overflows can be
 * avoided. Note that regardless of whether the output string has to
 * be truncated to fit within the specified buffer size, unless the
 * buffer size is specified as being 0 or less, the output string is
 * always terminated with a '\0' character within the specified size
 * limit.
 *
 * Input:
 *  buffer     char *  The buffer in which to format the output string.
 *  size     size_t    The allocated size of buffer[].
 *  fmt  const char *  The printf-style format.
 *  ...                The arguments refered to by fmt[].
 * Output:
 *  return      int    The number of bytes that the format attempted to
 *                     write to buffer[]. This differs from the actual
 *                     number written when the limit imposed by the
 *                     'size' argument causes the output string to be
 *                     truncated. On error, -1 is returned.
 */
int ccb_snprintf(char *buffer, size_t size, const char *fmt, ...)
#ifdef __GNUC__                          /* Arrange for gcc to check the */
__attribute__ ((format (printf, 3, 4))); /* printf-style format and arguments */
#else
 ;
#endif


/*.......................................................................
 * Provide an portable equivalent of vsnprint(). This operates identically
 * to ccb_snprintf(), except that instead of a variable argument list,
 * it takes a single va_list argument list.
 *
 * Input:
 *  buffer     char *  The buffer in which to format the output string.
 *  size     size_t    The allocated size of buffer[].
 *  fmt  const char *  The printf-style format.
 *  ap      va_list    The list of arguments refered to by fmt[].
 * Output:
 *  return      int    The number of bytes that the format attempted to
 *                     write to buffer[]. This differs from the actual
 *                     number written when the limit imposed by the
 *                     'size' argument causes the output string to be
 *                     truncated. On error, -1 is returned.
 */
int ccb_vsnprintf(char *buffer, size_t size, const char *fmt, va_list ap);

#endif
