#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "ccbqueue.h"
#include "freelist.h"

/*
 * Define a message queue list node.
 */
typedef struct CCBMQueueNode CCBMQueueNode;
struct CCBMQueueNode {
  CCBMQueueNode *next;    /* The next (newer) node in the list */
  void *msg;              /* The queued message */
};

/*
 * Objects of the following type are used to encapsulate queues
 * of messages.
 */
struct CCBMessageQueue {
  FreeList *msg_mem;        /* A free-list of message buckets */
  FreeList *node_mem;       /* A free-list of CCBMQueueNode nodes */
  CCBMQueueNode *head;      /* The head of the list (oldest queued message) */
  CCBMQueueNode *tail;      /* The tail of the list (newest queued message) */
  size_t qwidth;            /* The size of each message bucket */
  size_t qlength;           /* The length of the queue */
  size_t nqueued;           /* The number of messages in the queue */
  CCBMQFullAction mode;     /* What to do if a message is received while */
                            /*  the queue is full. */
  int draining;             /* True while draining a full queue (only set */
                            /*  when mode==CCB_MQ_DRAIN). */
};


/*.......................................................................
 * Create a new CCBMessageQueue object.
 *
 * Input:
 *  qwidth           size_t    The maximum number of bytes in each message.
 *  qlength          size_t    THe maximum number of bytes in the queue.
 *  mode    CCBMQFullAction    Specify what to do if a new message is
 *                             received when the queue is full.
 * Output:
 *  return  CCBMessageQueue *  The new object, or NULL on error.
 */
CCBMessageQueue *new_CCBMessageQueue(size_t qwidth, size_t qlength,
				     CCBMQFullAction mode)
{
  CCBMessageQueue *mq;  /* The object to be returned */
/*
 * Check the arguments.
 */
  if(qwidth < 1 || qlength < 1) {
    errno = EINVAL;
    return NULL;
  };
/*
 * Allocate the container.
 */
  mq = malloc(sizeof(CCBMessageQueue));
  if(!mq) {
    errno = ENOMEM;
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBMessageQueue().
 */
  mq->msg_mem = NULL;
  mq->node_mem = NULL;
  mq->head = NULL;
  mq->tail = NULL;
  mq->qwidth = qwidth;
  mq->qlength = qlength;
  mq->nqueued = 0;
  mq->mode = mode;
  mq->draining = 0;
/*
 * Allocate the free-list of message buckets.
 */
  mq->msg_mem = _new_FreeList(qwidth, qlength);
  if(!mq->msg_mem)
    return del_CCBMessageQueue(mq);
/*
 * Allocate the free-list of message-queue nodes.
 */
  mq->node_mem = _new_FreeList(sizeof(CCBMQueueNode), qlength);
  if(!mq->node_mem)
    return del_CCBMessageQueue(mq);
  return mq;
}

/*.......................................................................
 * Delete a CCBMessageQueue object.
 *
 * Input:
 *  mq     CCBMessageQueue *  The object to be deleted.
 * Output:
 *  return CCBMessageQueue *  The deleted object (always NULL).
 */
CCBMessageQueue *del_CCBMessageQueue(CCBMessageQueue *mq)
{
  if(mq) {
/*
 * Delete the free-lists. Note that this deletes any nodes and
 * messages that are currently in the message queue.
 */
    mq->msg_mem = _del_FreeList(mq->msg_mem, 1);
    mq->node_mem = _del_FreeList(mq->node_mem, 1);
/*
 * The message queue was deleted by deleting the free-lists.
 */
    mq->head = mq->tail = NULL;
    mq->nqueued = 0;
/*
 * Delete the container.
 */
    free(mq);
  };
  return NULL;
}

/*.......................................................................
 * Add a new message to the queue by copying the first 'size' bytes of
 * 'data' into a new queue bucket, then prepending this bucket to the
 * queue.  The return value is the number of bytes copied from *data
 * into the bucket. If the queue is full, this will be zero. Otherwise
 * it will be the minimum of the value of the 'size' argument of this
 * function, and the qwidth argument of new_CCBMessageQueue().
 *
 * Input:
 *  mq     CCBMessageQueue *  The queue to add the message to.
 *  data              void *  The message to be recorded.
 *  size            size_t    The number of bytes in the message
 *                            that is pointed to by 'data'.
 * Output:
 *  return          size_t    The number of bytes of '*data' that
 *                            were added to the queue (see the
 *                            explanation above).
 */
int ccb_mq_push(CCBMessageQueue *mq, void *data, size_t size)
{
  CCBMQueueNode *node = NULL;   /* The node of the newly queued message */
/*
 * Check the arguments.
 */
  if(!mq || (!data && size!=0)) {
    errno = EINVAL;
    return 0;
  };
/*
 * Is the queue either full or draining?
 */
  if(mq->nqueued >= mq->qlength || mq->draining) {
    switch(mq->mode) {
    case CCB_MQ_DRAIN:   /* Reject messages until the queue has been emptied */
      mq->draining = 1;
      errno = EAGAIN;
      return 0;
      break;
    case CCB_MQ_WRAP:    /* Discard the oldest message to make room */
      ccb_mq_pop(mq, NULL, 0);
      break;
    case CCB_MQ_DROP:    /* Simply reject the message */
    default:
      errno = EAGAIN;
      return 0;
      break;
    };
  };
/*
 * Get a message queue node to store the message in.
 */
  node = _new_FreeListNode(mq->node_mem);
  if(!node)
    return 0;
/*
 * Get the bucket in which to copy the message.
 */
  node->msg = _new_FreeListNode(mq->msg_mem);
  if(!node->msg) {
    node = _del_FreeListNode(mq->node_mem, node);
    return 0;
  };
/*
 * If the size of the message exceeds the size of a bucket, adjust
 * it to fit.
 */
  if(size > mq->qwidth)
    size = mq->qwidth;
/*
 * Copy the message into the bucket.
 */
  if(size > 0)
    memcpy(node->msg, data, size);
/*
 * Append the message to the queue.
 */
  node->next = NULL;
  if(mq->tail) {
    mq->tail->next = node;
    mq->tail = node;
  } else {
    mq->head = mq->tail = node;
  };
  mq->nqueued++;
/*
 * Return an indication of the number of bytes that were stored.
 */
  return size;
}

/*.......................................................................
 * Remove the oldest message from the queue, and return a copy of it
 * in the object pointed to by 'data', which has room for up to 'size'
 * bytes. The return value is the number of bytes copied into *data.
 * This will be zero if the queue is empty, or the minimum of the
 * value of the 'size' argument of this function, and the 'qwidth'
 * argument of new_CCBMessageQueue(). Any bytes that won't fit into
 * the return argument are silently discarded.
 *
 * Input:
 *  mq     CCBMessageQueue *  The queue to remove a message from.
 *  data              void *  The removed message will be stored in
 *                            *data.
 *  size            size_t    The maximum number of bytes to copy
 *                            into *data.
 * Output:
 *  return          size_t    The number of bytes of '*data' that
 *                            were copied into *data (see the
 *                            explanation above).
 */
int ccb_mq_pop(CCBMessageQueue *mq, void *data, size_t size)
{
  CCBMQueueNode *node;   /* The node of the removed message */
/*
 * Check the arguments.
 */
  if(!mq || (!data && size!=0)) {
    errno = EINVAL;
    return 0;
  };
/*
 * Is the queue empty?
 */
  if(mq->nqueued < 1) {
    errno = EAGAIN;
    return 0;
  };
/*
 * Remove the oldest message from the head of the list.
 */
  node = mq->head;
  mq->head = node->next;
  node->next = NULL;
  if(!mq->head)
    mq->tail = NULL;
  mq->nqueued--;
/*
 * How many bytes of the message can be copied into *data?
 */
  if(size > mq->qwidth)
    size = mq->qwidth;
/*
 * Copy up to 'size' bytes of the message into *data.
 */
  if(size > 0)
    memcpy(data, node->msg, size);
/*
 * Return the message container to the freelist.
 */
  node->msg = _del_FreeListNode(mq->msg_mem, node->msg);
/*
 * Return the node to the freelist.
 */
  node = _del_FreeListNode(mq->node_mem, node);
/*
 * An empty queue no longer has any messages to be drained.
 */
  if(mq->nqueued == 0)
    mq->draining = 0;
/*
 * Return an indication of the number of bytes that were retrieved.
 */
  return size;
}

/*.......................................................................
 * Empty the queue by throwing away all currently queued messages.
 *
 * Input:
 *  mq     CCBMessageQueue *  The queue to remove a message from.
 */
void ccb_mq_purge(CCBMessageQueue *mq)
{
  if(mq) {
/*
 * Simply return all queue nodes and buckets to their respective
 * free-lists.
 */
    _rst_FreeList(mq->node_mem);
    _rst_FreeList(mq->msg_mem);
/*
 * The queue is now empty.
 */
    mq->head = mq->tail = NULL;
    mq->nqueued = 0;
    mq->draining = 0;
  };
}

/*.......................................................................
 * Query how many more messages can currently be queued.
 *
 * Input:
 *  mq     CCBMessageQueue *  The queue to remove a message from.
 * Output:
 *  return          size_t    The number of messages that can be added
 *                            to the queue before it becomes full.
 */
size_t ccb_mq_space(CCBMessageQueue *mq)
{
  if(mq) {
    if(mq->draining)  /* No messages are accepted while the queue drains */
      return 0;
    return mq->qlength - mq->nqueued;
  };
  return 0;
}

/*.......................................................................
 * Query how many messages are currently queued.
 *
 * Input:
 *  mq     CCBMessageQueue *  The queue to query.
 * Output:
 *  return          size_t    The number of messages that are currently
 *                            in the queue.
 */
size_t ccb_mq_occupancy(CCBMessageQueue *mq)
{
  return mq ? mq->nqueued : 0;
}

/*.......................................................................
 * Return an indication of whether the queue is currently closed while
 * draining old messages (only if the mode argument of
 * new_CCBMessageQueue() was specified as CCB_MQ_DRAIN).
 *
 * Input:
 *  mq     CCBMessageQueue *  The queue to query.
 * Output:
 *  return             int    0 - Queue open to new messages.
 *                            1 - Queue closed while old messages are
 *                                being drained.
 */
int ccb_mq_draining(CCBMessageQueue *mq)
{
  return !mq || mq->draining;
}
