#ifndef ccbtelemetrymsg_h
#define ccbtelemetrymsg_h

#include "netbuf.h"
#include "ccbnetobj.h"
#include "ccbconstants.h"

/*-----------------------------------------------------------------------*
 * This header defines the message data-structures and description       *
 * tables that are used to exchange telemetry messages between a CCB     *
 * manager and a CCB server. Note that although this is a header file,   *
 * the included message-description table and its contents are all       *
 * declared static. This prevents namespace polution while keeping all   *
 * definitions in one place. An alternative would be to place the        *
 * message data-structures in this header, but the corresponding         *
 * message-table description entries in a separate C file. This is       *
 * judged to be a recipe for errors, and since this header file is only  *
 * included by 2 C files, ccbtelemetryserver.c and ccbtelemetryclient.c, *
 * and the table is trivially small, the overhead of defining the table  *
 * anew wherever this header is included is judged to be insignificant.  *
 *-----------------------------------------------------------------------*/

#ifndef CCB_ARRAY_DIM
#define CCB_ARRAY_DIM(array) (sizeof(array)/sizeof(array[0]))
#endif

/*-----------------------------------------------------------------------
 * Define the enumerators and structures of telemetry messages sent by the
 * CCB server to the CCB manager.
 */
typedef enum {
  CCB_INTEG_MSG,       /* An integration data message */
  CCB_MONITOR_MSG,     /* A monitoring data message */
  CCB_LOG_MSG,         /* A log message */
  CCB_TELEM_PING_REPLY /* A reply to a ping command */
} CCBTelemetryType;

/*
 * The first member of all telemetry message structures is a
 * CCBTelemetryHeader structure.
 */
typedef struct {
  CCBNetMsg base;         /* The base-class of all messages */
  unsigned long mjd;      /* The MJD UTC day number */
  unsigned long sec;      /* The time of day (seconds since 0H UTC) */
  unsigned long ns;       /* The number of nano-seconds after 'sec' */
} CCBTelemetryHeader;

/*
 * Provide a macro to describe the contents of the telemetry header of
 * each message type.
 */
#define CCB_TELEMETRY_HEADER \
  {"base",  offsetof(CCBTelemetryHeader,base), CCB_NET_MSG, 1}, \
  {"mjd",   offsetof(CCBTelemetryHeader,mjd),  CCB_NET_ULONG, 1}, \
  {"sec",   offsetof(CCBTelemetryHeader,sec),  CCB_NET_ULONG, 1}, \
  {"ns",    offsetof(CCBTelemetryHeader,ns), CCB_NET_ULONG, 1},

/* An integrated data message */

typedef struct {
  CCBTelemetryHeader head;           /* head.base.type=CCB_INTEG_MSG */
  unsigned long scan;                /* The scan number */
  unsigned long id;                  /* The integration ID */
  unsigned short flags;              /* Integration status flags */
  unsigned long data[CCB_MAX_INTEG]; /* The integrated data */
} CCBIntegMsg;

static const CCBNetMsgMember ccb_integ_msg_members[] = {
  CCB_TELEMETRY_HEADER
  {"scan", offsetof(CCBIntegMsg,scan), CCB_NET_ULONG, 1},
  {"id",   offsetof(CCBIntegMsg,id),   CCB_NET_ULONG, 1},
  {"flags",offsetof(CCBIntegMsg,flags),CCB_NET_USHORT,1},
  {"data", offsetof(CCBIntegMsg,data), CCB_NET_ULONG, CCB_MAX_INTEG}
};

/* A monitor-data message */

typedef struct {
  CCBTelemetryHeader head;           /* head.base.type=CCB_MONITOR_MSG */
  unsigned long scan;                  /* The scan number */
  unsigned long id;                    /* The monitor ID */
  unsigned long data[CCB_MAX_MONITOR]; /* The monitor data */
} CCBMonitorMsg;

static const CCBNetMsgMember ccb_monitor_msg_members[] = {
  CCB_TELEMETRY_HEADER
  {"scan", offsetof(CCBMonitorMsg,scan), CCB_NET_ULONG, 1},
  {"id",   offsetof(CCBMonitorMsg,id),   CCB_NET_ULONG, 1},
  {"data", offsetof(CCBMonitorMsg,data), CCB_NET_ULONG, CCB_MAX_MONITOR}
};

/* A log message */

typedef struct {
  CCBTelemetryHeader head;         /* head.base.type=CCB_LOG_MSG */
  char msg[CCB_MAX_LOG];    /* The message to be logged */
  unsigned long id;                  /* A unique message identifier */
  unsigned short level;              /* The severity level of the message */
} CCBLogMsg;

static const CCBNetMsgMember ccb_log_msg_members[] = {
  CCB_TELEMETRY_HEADER
  {"msg",      offsetof(CCBLogMsg,msg),      CCB_NET_ASCII, CCB_MAX_LOG},
  {"id",       offsetof(CCBLogMsg,id),       CCB_NET_ULONG, 1},
  {"level",    offsetof(CCBLogMsg,level),    CCB_NET_USHORT, 1}
};

/* A reply to a ping command */

typedef struct {
  CCBTelemetryHeader head;    /* head.base.type=CCB_TELEM_PING_REPLY */
  unsigned long seq;          /* The sequence number of the ping */
                              /*  that is being replied to. */
} CCBTelemPingReply;

static const CCBNetMsgMember ccb_telem_ping_reply_members[] = {
  CCB_TELEMETRY_HEADER
  {"seq",      offsetof(CCBTelemPingReply,seq), CCB_NET_ULONG,  1},
};

/*
 * Create a union of the above message structures.
 */
typedef union {
  CCBTelemetryHeader head;     /* The common telemetry header */
  CCBIntegMsg integ;           /* An integrated-data message */
  CCBMonitorMsg monitor;       /* A monitor data message */
  CCBLogMsg log;               /* A log message */
  CCBTelemPingReply ping;      /* A reply to a ping command */
} CCBTelemetryMessage;

/*
 * Collect all of the object definitions in an array.
 */
static const CCBNetMsgInfo ccb_telemetry_structures[] = {
  {CCB_INTEG_MSG,      "integ",
   ccb_integ_msg_members, CCB_ARRAY_DIM(ccb_integ_msg_members),
   sizeof(CCBIntegMsg)},
  {CCB_MONITOR_MSG,    "monitor",
   ccb_monitor_msg_members, CCB_ARRAY_DIM(ccb_monitor_msg_members),
   sizeof(CCBMonitorMsg)},
  {CCB_LOG_MSG,        "log",
   ccb_log_msg_members, CCB_ARRAY_DIM(ccb_log_msg_members),
   sizeof(CCBLogMsg)},
  {CCB_TELEM_PING_REPLY, "ping-reply",
   ccb_telem_ping_reply_members, CCB_ARRAY_DIM(ccb_telem_ping_reply_members),
   sizeof(CCBTelemPingReply)},
};

/*
 * Form a table of message-descriptions for use by the functions in
 * ccbnetobj.h.
 */
static const CCBNetMsgTable ccb_telemetry_msg_table = {
  "telemetry", ccb_telemetry_structures,
  sizeof(ccb_telemetry_structures)/sizeof(CCBNetMsgInfo),
  sizeof(CCBTelemetryMessage)
};

#endif
