#include <limits.h>
#include <errno.h>
#include <string.h>

#include "ccbunsigned.h"

/*
 * Parameterize the number of bits in an unsigned long.
 */
#define ULONG_BITSIZE (sizeof(long) * CHAR_BIT)

/*
 * Half of the number of bit in an unsigned long.
 */
#define HALF_ULONG_BITSIZE (ULONG_BITSIZE/2)

static int value_of_digit(char c);
static int digit_of_value(int value);

/*.......................................................................
 * Set a specified bit within an extended unsigned integer stored in
 * an array of unsigned long elements.
 *
 * Input:
 *  a     unsigned long *  The array of 'na' unsigned long elements
 *                         to be modified. The zero'th element is
 *                         interpretted as being the least significant
 *                         element.
 *  na         unsigned    The number of elements in 'a'.
 *  bit        unsigned    The bit to set, in the range
 *                         0..(na*sizeof(*a) - 1).
 * Output:
 *  return          int    0 - OK.
 *                         1 - The specified bit is beyond the most
 *                             significant bit of 'a'.
 */
int ccb_ull_set_bit(unsigned long *a, unsigned na, unsigned bit)
{
/*
 * Which element of the array of unsigned long's is the bit in?
 */
  int ilong = bit / ULONG_BITSIZE;
/*
 * Is the specified bit is outside the bounds of the integer?
 */
  if(ilong >= na || !a) {
    errno = ERANGE;
    return 1;
  };
/*
 * Set the specified bit within the located element of 'a'.
 */
  a[ilong] |= 1U << (bit - ilong * ULONG_BITSIZE);
  return 0;
}

/*.......................................................................
 * Clear a specified bit within an extended unsigned integer stored in
 * an array of unsigned long elements.
 *
 * Input:
 *  a     unsigned long *  The array of 'na' unsigned long elements
 *                         to be modified. The zero'th element is
 *                         interpretted as being the least significant
 *                         element.
 *  na         unsigned    The number of elements in 'a'.
 *  bit        unsigned    The bit to clear, in the range
 *                         0..(na*sizeof(*a) - 1).
 */
void ccb_ull_clear_bit(unsigned long *a, unsigned na, unsigned bit)
{
/*
 * Which element of the array of unsigned long's is the bit in?
 */
  int ilong = bit / ULONG_BITSIZE;
/*
 * Clear the bit if it is within the bounds of 'a'. Otherwise, by
 * definition, the target bit is already zero, since it is beyond
 * the most significant bit in the integer.
 */
  if(ilong < na && a)
    a[ilong] &= ~(1U << (bit - ilong * ULONG_BITSIZE));
  return;
}

/*.......................................................................
 * Get the value of a specified bit within an extended unsigned integer
 * stored in an array of unsigned long elements.
 *
 * Input:
 *  a  const unsigned long *  The array of 'na' unsigned long elements
 *                            to be shifted. The zero'th element is
 *                            interpretted as being the least significant
 *                            element.
 *  na            unsigned    The number of elements in 'a'.
 *  bit           unsigned    The bit to query, in the range
 *                            0..(na*sizeof(*a) - 1).
 * Output:
 *  return             int    The value of the bit (0 or 1).
 */
int ccb_ull_get_bit(const unsigned long *a, unsigned na, unsigned bit)
{
/*
 * Which element of the array of unsigned long's is the bit in?
 */
  int ilong = bit / ULONG_BITSIZE;
/*
 * If the specified bit is above the most significant bit of the integer,
 * then the result is clearly zero.
 */
  if(ilong >= na || !a)
    return 0;
/*
 * Get the specified bit from within the located element of 'a'.
 */
  return (a[ilong] >> (bit - ilong * ULONG_BITSIZE)) & 1U;
}

/*.......................................................................
 * Perform an arithmetic shift left on a large unsigned integer
 * recorded as an array of unsigned long elements.
 *
 * Input/Output:
 *  a     unsigned long *  The array of 'na' unsigned long elements
 *                         to be shifted. The zero'th element is
 *                         interpretted as being the least significant
 *                         element.
 * Input:
 *  na         unsigned    The number of elements in 'a'.
 *  nbits      unsigned    The size of the shift in bits.
 */
void ccb_ull_shift_left(unsigned long *a, unsigned na, unsigned nbits)
{
/*
 * How many complete elements does the shift correspond to?
 */
  int nlong = nbits / ULONG_BITSIZE;
  int i;
/*
 * If shifting by >= the width of the array of longs, the result is zero.
 */
  if(nlong >= na || !a) {
    ccb_ull_assign_zero(a, na);
  } else {
/*
 * Work out the number of remaining bits to shift after
 * shifting by 'nlong' elements.
 */
    unsigned shift_remainder = (nbits - ULONG_BITSIZE * (unsigned) nlong);
/*
 * Perform the shift, working from the most significant element of
 * the array to the element that will have bits of the least
 * significant element shifted into it.
 */
    for(i=na-1; i>=nlong; i--) {
      a[i] = a[i-nlong] << shift_remainder;
      if(i-nlong-1 >= 0 && shift_remainder != 0)
	a[i] |= a[i-nlong-1] >> (ULONG_BITSIZE - shift_remainder);
    };
/*
 * The remaining least significant elements are all reduced to zero.
 */
    ccb_ull_assign_zero(a, nlong);
  };
  return;
}

/*.......................................................................
 * Perform an arithmetic shift right on a large unsigned integer
 * implemented by an array of unsigned long elements.
 *
 * Input/Output:
 *  a     unsigned long *  The array of 'na' unsigned long elements
 *                         to be shifted. The zero'th element is
 *                         interpretted as being the least significant
 *                         element.
 * Input:
 *  na         unsigned    The number of elements in 'a'.
 *  nbits      unsigned    The size of the shift in bits.
 */
void ccb_ull_shift_right(unsigned long *a, unsigned na, unsigned nbits)
{
/*
 * How many complete elements does the shift correspond to?
 */
  int nlong = nbits / ULONG_BITSIZE;
  int i;
/*
 * If shifting by >= the width of an extended unsigned long,
 * the result is zero.
 */
  if(nlong >= na || !a) {
    ccb_ull_assign_zero(a, na);
  } else {
/*
 * Work out the number of remaining bits to shift after
 * shifting by 'nlong' elements.
 */
    unsigned shift_remainder = (nbits - ULONG_BITSIZE * (unsigned) nlong);
/*
 * Perform the shift, working from the least significant element of
 * the array to the element that will have bits of the most
 * significant element shifted into it.
 */
    for(i=0; i<na-nlong; i++) {
      a[i] = a[i+nlong] >> shift_remainder;
      if(i+nlong+1 < na && shift_remainder != 0)
	a[i] |= a[i+nlong+1] << (ULONG_BITSIZE - shift_remainder);
    };
/*
 * The remaining most significant elements are all reduced to zero.
 */
    ccb_ull_assign_zero(a + (na-nlong), nlong);
  };
  return;
}

/*.......................................................................
 * Assign zeros to specified elements of an array of unsigned longs.
 *
 * Input/Output:
 *  a      unsigned long *  The first element of the sub-array to be
 *                          zeroed.
 * Input:
 *  na          unsigned    The number of elements to be zeroed.
 */
void ccb_ull_assign_zero(unsigned long *a, unsigned na)
{
  int i;
  if(a) {
    for(i=0; i<na; i++)
      a[i] = 0;
  };
  return;
}

/*.......................................................................
 * Compare two extended unsigned long integers.
 *
 * Input:
 *  a  const unsigned long *  The array of 'na' unsigned longs which
 *                            constitute the integer left-hand term of
 *                            the comparison.
 *  na            unsigned    The number of elements in 'a'.
 *  b  const unsigned long *  The array of 'nb' unsigned longs which
 *                            constitute the integer right-hand term of
 *                            the comparison.
 *  nb            unsigned    The number of elements in 'b'.
 * Output:
 *  return             int    -1: a<b
 *                             0: a==b
 *                             1: a>b
 */
int ccb_ull_compare(const unsigned long *a, unsigned na,
		    const unsigned long *b, unsigned nb)
{
  unsigned ncom;  /* The number of elements in common to the two integers */
  int i;
/*
 * Enforce consistency between the number of elements claimed and whether
 * arrays have been provided.
 */
  if(!a && na)
    na = 0;
  if(!b && nb)
    nb = 0;
/*
 * If 'a' contains more sigificant elements than 'b', and any of these
 * elements are non-zero, then a>b.
 */
  if(na > nb) {
    for(i=na-1; i>nb-1; i--) {
      if(a[i])
	return 1;             /* a>b */
    };
    ncom = nb;
/*
 * If 'b' contains more sigificant elements than 'a', and any of these
 * elements are non-zero, then a<b.
 */
  } else if(nb > na) {
    for(i=nb-1; i>na-1; i--) {
      if(b[i])
	return -1;            /* a<b */
    };
    ncom = na;
/*
 * na==nb.
 */
  } else {
    ncom = na;
  };
/*
 * Compare the remaining overlapping 'ncom' least significant
 * elements, working from the most significant to the least
 * significant elements.
 */
  for(i=ncom-1; i>=0; i--) {
    if(a[i] < b[i])
      return -1;              /* a<b */
    else if(a[i] > b[i])
      return 1;               /* a>b */
  };
/*
 * The integers are equal.
 */
  return 0;
}

/*.......................................................................
 * Make a copy of an extended unsigned long.
 *
 * Input:
 *  a const unsigned long *  The array of 'na' unsigned longs to be copied.
 *  na           unsigned    The number of elements in 'a'.
 * Input/Output:
 *  b       unsigned long *  The array of 'nb' unsigned longs into which to
 *                           copy 'a'.
 * Input:
 *  nb           unsigned    The number of elements in 'b'. Note that it is
 *                           ok for nb < na, but only if elements of a[]
 *                           with indexes > na-1 are all zero.
 * Output:
 *  return            int    0 - OK.
 *                           1 - There isn't enough room in 'b' to make
 *                               a copy of 'a'.
 */
int ccb_ull_copy(const unsigned long *a, unsigned na,
		 unsigned long *b, unsigned nb)
{
  int i;
/*
 * Enforce consistency between the number of elements claimed and whether
 * arrays have been provided.
 */
  if(!a && na)
    na = 0;
  if(!b && nb)
    nb = 0;
/*
 * Copy the elements of 'a' to 'b'.
 */
  for(i=0; i<na; i++) {
    if(i < nb) {
      b[i] = a[i];
    } else if(a[i]) {
      errno = ERANGE;
      return 1;
    };
  };
/*
 * If 'a' is smaller than 'b', pad 'b' with zeroes.
 */
  for( ; i<nb; i++)
    b[i] = 0;
  return 0;
}

/*.......................................................................
 * Return true (non-zero) if an extended unsigned long integer is zero.
 *
 * Input:
 *  a  const unsigned long *  The extended integer to be examined.
 *  na            unsigned    The number of elements in 'a'.
 * Output:
 *  return             int     0 - (a is zero)
 *                            !0 - (a is non-zero)
 */
int ccb_ull_is_zero(const unsigned long *a, unsigned na)
{
  int i;
/*
 * Enforce consistency between the number of elements claimed and
 * whether an array has been provided.
 */
  if(!a && na)
    na = 0;
/*
 * If any element of 'a' is non-zero, then the extended integer that
 * they compose, is also non-zero.
 */
  for(i=0; i<na; i++) {
    if(a[i])
      return 0;
  };
  return 1;    /* To get here, all components of 'a' must be zero */
}

/*.......................................................................
 * Return the zero-relative index of the most-significant non-zero bit
 * within an extended unsigned long integer.
 *
 * Input:
 *  a  const unsigned long *  The extended integer to be examined.
 *  na            unsigned    The number of elements in 'a'.
 * Output:
 *  return             int    The index of the most significant non-zero
 *                            bit, or -1 if ull is zero.
 */
int ccb_ull_msb(const unsigned long *a, unsigned na)
{
  unsigned long top;   /* The most significant non-zero element of 'a' */
  int shift;           /* The size of a bit shift */
  int i;
/*
 * Enforce consistency between the number of elements claimed and
 * whether an array has been provided.
 */
  if(!a && na)
    na = 0;
/*
 * Find the most significant non-zero element in 'a'.
 */
  for(i=na-1; i>=0 && a[i]==0; i--)
    ;
/*
 * If all elements are zero, there is no most-significant non-zero bit.
 */
  if(i<0)
    return -1;
/*
 * Get a copy of the most significant non-zero element, so that we
 * can shift it without affecting the original.
 */
  top = a[i];
/*
 * Repeatedly shift the element 1-bit right until it becomes zero.
 */
  for(shift=1; (top >>= 1U) != 0; shift++)
    ;
  return shift-1 + ULONG_BITSIZE * i;
}

/*.......................................................................
 * Subtract one extended unsigned long integer from another.
 *
 * Input:
 *  a  const unsigned long *  The array of 'na' unsigned longs which
 *                            constitute the integer from which to
 *                            subtract.
 *  na            unsigned    The number of elements in 'a'.
 *  b  const unsigned long *  The array of 'nb' unsigned longs which
 *                            constitute the integer to be subtracted.
 *  nb            unsigned    The number of elements in 'b'.
 * Input/Output:
 *  dif      unsigned long *  The result of the subtraction. Note that
 *                            it is okay for dif[] and a[] to be the same
 *                            array.
 * Input:
 *  ndif          unsigned    The number of elements in 'dif'.
 * Output:
 *  return             int    0 - OK.
 *                            1 - a-b would be negative, so the
 *                                subtraction could not be performed.
 */
int ccb_ull_subtract(const unsigned long *a, unsigned na,
		     const unsigned long *b, unsigned nb,
		     unsigned long *dif, unsigned ndif)
{
  unsigned borrow = 0;  /* The amount to borrow in the next iteration of the */
                        /*  subtraction */
  int i;
/*
 * Get the size of the largest of the two input integers.
 */
  unsigned nmax = na > nb ? na : nb;
/*
 * Enforce consistency between the number of elements claimed and whether
 * arrays have been provided.
 */
  if(!a && na)
    na = 0;
  if(!b && nb)
    nb = 0;
  if(!dif && ndif)
    ndif = 0;
/*
 * If a < b, the result is negative and can't be represented in the
 * returned unsigned integer (we could use modulo unsigned arithmetic,
 * as in C, but it is more useful to flag this as an error condition).
 */
  if(ccb_ull_compare(a, na, b, nb) < 0) {
    errno = ERANGE;
    return 1;
  };
/*
 * Working from the least significant to the most significant elements,
 * subtract elements of 'a' from 'b', leaving the result in 'a'.
 */
  for(i=0; i<nmax; i++) {
    unsigned long ai = i<na ? a[i] : 0;
    unsigned long bi = i<nb ? b[i] : 0;
/*
 * If the previous iteration of this loop requested that 1 be borrowed
 * from the following iteration, subtract 1 from 'a'. If this can't
 * be done without making a[i] negative, postpone the borrow operation
 * to the next iteration.
 */
    if(borrow) {
/*
 * If the borrow operation can be completed in this iteration,
 * clear the borrow flag. Otherwise leave it set to request a borrow
 * from the next iteration.
 */
      if(ai > 0)
	borrow = 0;
/*
 * Subtract the borrowed 1 from 'a'. Note that because of the modulo
 * modulo-datatype-size rules of unsigned arithmetic in C, if 'ai' is
 * negative, ai-- is equivalent to (1<<bit_width_of_ai - 1). This is
 * what we want, since in this case we are borrowing
 * (1<<bit_width_of_ai) from the next most significant element of 'a',
 * to make 'ai' big enough to subtract 1 from in this iteration.
 */
      ai--;
    };
/*
 * If the latest element of 'a' is less than that of 'b', flag the
 * need to borrow 1 from the next most significant element of 'a'.
 */
    if(ai < bi)
      borrow = 1;
/*
 * Subtract the current element of b[] from the same element of a[].
 * Note that under the modulo-datatype-size rules of unsigned
 * arithmetic in C, (0 - x) is equivalent to (1<<bit_width_of_datatype - x).
 * This means that if a[i]<b[i], subtracting b[i] from a[i] is equivalent
 * to subtracting (1<<bit_width_of_long - b[i]) from a[i], which is what
 * we want, since in this case we have just borrowed 1<<bit_width_of_long
 * from the next most significant element.
 *
 * Is there space in dif[] to store the result?
 */
    if(i<ndif) {
      dif[i] = ai - bi;     /* Store the result */
    } else if(bi != ai) { /* Is the result non-zero? */
      errno = ERANGE;
      return 1;           /* There is no room to store the non-zero result */
    };
  };
  return 0;
}

/*.......................................................................
 * Add one extended unsigned long integer to another.
 *
 * Input:
 *  a  const unsigned long *  The array of 'na' unsigned longs which
 *                            constitute the first of the integers to be
 *                            added.
 *  na            unsigned    The number of elements in 'a'.
 *  b  const unsigned long *  The array of 'nb' unsigned longs which
 *                            constitute the second of the integers to be
 *                            added.
 *  nb            unsigned    The number of elements in 'b'.
 * Input/Output:
 *  sum      unsigned long *  The result of the addition (note that a[]
 *                            and sum[] can be the same array).
 * Input:
 *  nsum          unsigned    The number of elements in 'sum'.
 * Output:
 *  return            int     0 - OK.
 *                            1 - The addition overflowed the available
 *                                space in 'sum'.
 */
int ccb_ull_add(const unsigned long *a, unsigned na,
		const unsigned long *b, unsigned nb,
		unsigned long *sum, unsigned nsum)
{
  unsigned long sum_ai_bi; /* The sum a[i] + b[i] (excluding any carry bit) */
  unsigned long carry = 0; /* The overflow bit, if any from adding x to y */
  unsigned top_bit_sum;    /* The sum of the top-most bits during sums */
  int i;
/*
 * Get an unsigned long element whose top-most bit is the only bit set.
 */
  const unsigned long topbit = 1U << (ULONG_BITSIZE-1);
/*
 * Determine the size of the longest of the two input integers.
 */
  unsigned nmax = na > nb ? na : nb;
/*
 * Enforce consistency between the number of elements claimed and whether
 * arrays have been provided.
 */
  if(!a && na)
    na = 0;
  if(!b && nb)
    nb = 0;
  if(!sum && nsum)
    nsum = 0;
/*
 * Add first the least significant longs of the two numbers, then
 * the most significant longs of the two numbers, in the second
 * case also adding in any overflow from the first addition.
 */
  for(i=0; i<nmax; i++) {
    unsigned long ai = i<na ? a[i] : 0;
    unsigned long bi = i<nb ? b[i] : 0;
/*
 * To prevent the addition of ai and bi from overflowing, zero their
 * top-most bits, keeping a record of the sum of their value to be
 * added back later.
 */
    top_bit_sum = 0;
    if(ai & topbit) {
      ai &= ~topbit;
      top_bit_sum++;
    };
    if(bi & topbit) {
      bi &= ~topbit;
      top_bit_sum++;
    };
/*
 * Now that both top-most bits are known to be zero, it is safe to add
 * any pending carry to either ai (or bi).
 */
    if(carry) {
      ai++;
      carry = 0;
    };
/*
 * Add the two possibly-modified elements.
 */
    sum_ai_bi = ai + bi;
/*
 * Get the sum of the top-most bits that were removed, and the top-most
 * bit of the sum of ai and bi.
 */
    if(sum_ai_bi & topbit)
      top_bit_sum++;
/*
 * If top_bit_sum > 1, then we need to distribute it between the top-most
 * bit of the current sum and the carry bit to be added in the next
 * iteration. The carry bit and the top-most bit of the sum
 * effectively form a two bit integer, whose value should be the value
 * of top_bit_sum.
 */
    if(top_bit_sum & 1U)
      sum_ai_bi |= topbit;
    else
      sum_ai_bi &= ~topbit;
    carry = (top_bit_sum & 2U) == 2U;
/*
 * Is there space in sum[] to store the result?
 */
    if(i<nsum) {
      sum[i] = sum_ai_bi;  /* Store the result */
    } else if(sum_ai_bi > 0) {
      errno = ERANGE;
      return 1; /* The non-zero result can't be recoded, so sum[] overflowed */
    };
  };
/*
 * Did the final sum generate a carry bit?
 */
  if(carry) {
/*
 * Would adding the carry into the next element of sum[], overflow the
 * capacity of sum[]?
 */
    if(i >= nsum) {
      errno = ERANGE;
      return 1;
    };
/*
 * Assign the carry bit to the next most-significant element of sum[].
 */
    sum[i++] = 1;
  };
/*
 * Zero any remaining higher-significance elements in sum[].
 */
  while(i<nsum)
    sum[i++] = 0;
  return 0;
}

/*.......................................................................
 * Divide one extended unsigned long integer by another. Note that both
 * of the input integers are modified by this function.
 *
 * Input:
 *  a  const unsigned long *  The array of 'na' unsigned longs which
 *                            constitute the integer number to be divided.
 *  na            unsigned    The number of elements in 'a'.
 *  b  const unsigned long *  The array of 'nb' unsigned longs which
 *                            constitute the integer number to divide by.
 *  nb            unsigned    The number of elements in 'b'.
 * Input/Output:
 *  quot     unsigned long *  The quotient a/b. Note that quot[] and
 *                            b[] can be the same array.
 * Input:
 *  nquot         unsigned    The number of elements in 'quot'. This
 *                            should ideally be >= na.
 * Input/Output:
 *  rem      unsigned long *  The remainder of a/b. Note that rem[] and
 *                            a[] can be the same array.
 *  nrem          unsigned    The number of elements in rem[]. Ideally
 *                            this should be >= na, because internally
 *                            this array initially holds a copy of 'a'.
 * Output:
 *  return             int     0 - OK.
 *                             1 - Either 'b' is zero, or one or both
 *                                 of 'quot', 'rem' weren't big enough.
 */
int ccb_ull_divide(const unsigned long *a, unsigned na,
		   const unsigned long *b, unsigned nb,
		   unsigned long *quot, unsigned nquot,
		   unsigned long *rem, unsigned nrem)
{
  int i;
/*
 * Enforce consistency between the number of elements claimed and whether
 * arrays have been provided.
 */
  if(!a && na)
    na = 0;
  if(!b && nb)
    nb = 0;
  if(!quot && nquot)
    nquot = 0;
  if(!rem && nrem)
    nrem = 0;
/*
 * Reject division by zero.
 */
  if(ccb_ull_is_zero(b, nb)) {
    errno = EDOM;
    return 1;
  };
/*
 * Start with the remainder equal to the entirety of 'a'. As the
 * division procedes, multiples of the divisor will be subtracted from
 * the remainder, and their multipliers added to the quotient.
 */
  if(ccb_ull_copy(a, na, rem, nrem))
    return 1;
/*
 * If a<b, then the division is zero, and the remainder remains equal
 * to the entirety of 'a'.
 */
  if(ccb_ull_compare(a, na, b, nb) < 0) {
    ccb_ull_assign_zero(quot, nquot);
  } else {
/*
 * Determine the index of the most significant non-zero bit in the divisor.
 */
    int b_topbit = ccb_ull_msb(b, nb);
/*
 * Also determine the index of the most significant non-zero bit in the
 * numerator.
 */
    int a_topbit = ccb_ull_msb(a, na);
/*
 * Determine how many bits one would need to shift the divisor
 * left, to place its most-significant non-zero bit in the same location
 * as that of the number being divided.
 */
    int shift =  a_topbit - b_topbit;
/*
 * Copy the divisor into the quotient array, and shift the copy
 * left such that its highest non-zero bit is lined up with that
 * of the number being divided. As this gets shifted back one
 * place at a time, the vacated bit will be replaced by 1 if
 * the shifted divisor was subtracted from the remainder, or
 * 0 otherwise. This thus builds up a shifted version of the
 * final quotient.
 */
    ccb_ull_copy(b, nb, quot, nquot);
    ccb_ull_shift_left(quot, nquot, shift);
/*
 * With the divisor multiplied by successively smaller powers of 2,
 * starting with that the multiplied value now in 'quot', if the
 * result is less than the current value of the remainder, set the bit
 * that represents the current power of 2 multiplier, within the
 * output integer and subtract the multiplied divisor from 'rem'.
 */
    for(i=a_topbit/ULONG_BITSIZE; shift>=0; i--) {
/*
 * In order to allow quot[] to contain the copy of the scaled version
 * of the divisor in its low-order bits, while we are accumulating a
 * shifted version of the the quotient in the higher-order bits,
 * accumulate each long element of the shifted quotient first in the
 * temporary variable ctmp, recording it in quot[] each time that the
 * scaled version of the divisor vacates another element on being
 * shifted right.
 */
      unsigned long ctmp = 0;
/*
 * Work out element 'i' of the shifted quotient.
 */
      for( ; (b_topbit + shift)/ULONG_BITSIZE==i && shift>=0; shift--) {
/*
 * Can we subtract the latest power-of-two multiple of the divisor
 * from the remainder?
 */
	if(ccb_ull_compare(rem, nrem, quot, i+1) >= 0) {
	  ccb_ull_subtract(rem, nrem, quot, i+1, rem, nrem);
	  ctmp |= 1<<((b_topbit + shift) - i*ULONG_BITSIZE);
	};
/*
 * Divide the scaled divisor by two.
 */
	ccb_ull_shift_right(quot, i+1, 1);
      };
/*
 * Record the latest element of the quotient in the element that was
 * just vacated by shifting the scaled version of the divisor right.
 */
      quot[(b_topbit + shift + 1)/ULONG_BITSIZE] = ctmp;
    };
/*
 * Unshift the shifted quotient.
 */
    ccb_ull_shift_right(quot, nquot, b_topbit);
  };
  return 0;
}

/*.......................................................................
 * Multiply one extended unsigned long integer by another.
 *
 * Input:
 *  a  const unsigned long *  The array of 'na' unsigned longs which
 *                            constitute the first of the integer numbers
 *                            to be multiplied.
 *  na            unsigned    The number of elements in 'a'.
 *  b  const unsigned long *  The array of 'nb' unsigned longs which
 *                            constitute the second of the integer numbers
 *                            to multiplied.
 *  nb            unsigned    The number of elements in 'b'.
 * Input/Output:
 *  prod     unsigned long *  The product a*b.
 * Input:
 *  nprod         unsigned    The number of elements in 'prod'.
 *                            Note that if nprod is at least as large
 *                            as the sum of the number of non-zero
 *                            elements in a[] and b[], then the result
 *                            is guaranteed not to overflow.
 *  work     unsigned long *  A work array.
 *  nwork         unsigned    The number of elements in work[]. This
 *                            should be >= the smallest of na and nb.
 * Output:
 *  return             int     0 - OK.
 *                             1 - Insufficient room in 'prod'.
 */
int ccb_ull_multiply(const unsigned long *a, unsigned na,
		     const unsigned long *b, unsigned nb,
		     unsigned long *prod, unsigned nprod,
		     unsigned long *work, unsigned nwork)
{
  int a_topbit;  /* The most significant non-zero bit in 'a' */
  int b_topbit;  /* The most significant non-zero bit in 'b' */
  int bit;       /* The zero-relative index of a single bit */
/*
 * Enforce consistency between the number of elements claimed and whether
 * arrays have been provided.
 */
  if(!a && na)
    na = 0;
  if(!b && nb)
    nb = 0;
  if(!prod && nprod)
    nprod = 0;
/*
 * Zero-initialize the answer and the work array.
 */
  ccb_ull_assign_zero(prod, nprod);
  ccb_ull_assign_zero(work, nwork);
/*
 * If either of the terms being multiplied are zero, then so is
 * the result.
 */
  if(ccb_ull_is_zero(a, na) || ccb_ull_is_zero(b, nb))
    return 0;
/*
 * Sort 'a' and 'b', such that 'a' is the smaller of the two.
 */
  if(ccb_ull_compare(b, nb, a, na) < 0) {
    const unsigned long *tmp = a;
    unsigned ntmp = na;
    a = b;
    na = nb;
    b = tmp;
    nb = ntmp;
  };
/*
 * Determine the index of the most significant non-zero bits in 'a'
 * and 'b'.
 */
  a_topbit = ccb_ull_msb(a, na);
  b_topbit = ccb_ull_msb(b, nb);
/*
 * Copy 'b' into the work array and check that there will be room to shift
 * this copy left by a_topbit bits.
 */
  if(ccb_ull_copy(b, nb, work, nwork) ||
     a_topbit + b_topbit >= ULONG_BITSIZE * nwork) {
    errno = ERANGE;
    return 1;
  };
/*
 * Multiply 'b' by each non-zero bit of 'a', successively adding the
 * result of each of these multiplications to 'prod'.
 */
  for(bit=0; bit<=a_topbit; bit++) {
/*
 * If the latest power-of-two multiplier bit in 'a' is set, add
 * the correspondingly scaled version of 'b' (currently in 'work')
 * to the product.
 */
    if(ccb_ull_get_bit(a, na, bit)) {
      if(ccb_ull_add(prod, nprod, work, nwork, prod, nprod))
	return 1;
    };
/*
 * Get 'b' multiplied by the next highest power-of-two multiplier.
 */
    ccb_ull_shift_left(work, nwork, 1);
  };
  return 0;
}

/*.......................................................................
 * Read an extended unsigned integer from a string.
 *
 * Input:
 *  s         const char *  The string to read from.
 * Input/Output:
 *  a      unsigned long *  The place to record the integer.
 * Input:
 *  na          unsigned    The number of elements in 'a'.
 *  work1  unsigned long *  A work array.
 *  nwork1      unsigned    The number of elements in work1[]. This
 *                          should be >= na.
 *  work2  unsigned long *  A second work array.
 *  nwork2      unsigned    The number of elements in work2[]. This
 *                          should be >= na.
 * Input/Output:
 *  endp            char ** If endp!=NULL, a pointer to the next
 *                          unprocessed character will be left in *endp.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
int ccb_ull_input(const char *s, unsigned long *a, unsigned na,
		  unsigned long *work1, unsigned long nwork1,
		  unsigned long *work2, unsigned long nwork2,
		  char **endp)
{
  const unsigned long ten = 10;  /* Decimal multiplier */
  unsigned long n;               /* The value of the latest digit */
/*
 * Check the arguments.
 */
  if(!s)
    return 1;
/*
 * Zero initialize 'a'.
 */
  ccb_ull_assign_zero(a, na);
/*
 * Accumulate the answer in 'a'.
 */
  while((n=value_of_digit(*s)) != -1) {
/*
 * Multiply the currently accumulated value by 10.
 */
    if(ccb_ull_multiply(a, na, &ten, 1, work1, nwork1, work2, nwork2) ||
       ccb_ull_copy(work1, nwork1, a, na))
      return 1;
/*
 * Add in the new value, as the new least significant power-of-10
 * digit of 'a'.
 */
    if(ccb_ull_add(a, na, &n, 1, a, na))
      return 1;
/*
 * Prepare to read the next digit.
 */
    s++;
  };
/*
 * If requested, return a pointer to the next unprocessed digit.
 */
  if(endp)
    *endp = (char *)s;
  return 0;
}

/*.......................................................................
 * Convert a numeric digit to the value that it represents.
 *
 * Input:
 *  c      char   The digit to be translated.
 * Output:
 *  return  int   The value represented by the digit, or -1 if c doesn't
 *                contain a numeric digit.
 */
static int value_of_digit(char c)
{
/*
 * Lookup the value of the digit in c.
 */
  switch(c) {
  case '0':
    return 0;
  case '1':
    return 1;
  case '2':
    return 2;
  case '3':
    return 3;
  case '4':
    return 4;
  case '5':
    return 5;
  case '6':
    return 6;
  case '7':
    return 7;
  case '8':
    return 8;
  case '9':
    return 9;
  };
/*
 * Not a digit.
 */
  return -1;
}

/*.......................................................................
 * Write an extended unsigned integer to a string.
 *
 * Input:
 *  s                 char *  The string in which to write the integer.
 *  dim             size_t    The allocated size of s[].
 *  a  const unsigned long *  The integer to be written.
 *  na            unsigned    The number of elements in a[].
 *  work1    unsigned long *  A work array.
 *  nwork1        unsigned    The number of elements in work1[]. This
 *                            should be >= na.
 *  work2    unsigned long *  A second work array.
 *  nwork2        unsigned    The number of elements in work2[]. This
 *                            should be >= na.
 * Output:
 *  return             int    The number of characters written
 *                            (excluding the '\0'), or -1 on error.
 */
int ccb_ull_output(char *s, size_t dim, const unsigned long *a, unsigned na,
		   unsigned long *work1, unsigned nwork1,
		   unsigned long *work2, unsigned nwork2)
{
  const unsigned long ten = 10;  /* Decimal divisor */
  int nchar=0;                   /* The number of characters written so far */
/*
 * Check the arguments.
 */
  if(!s || dim < 2)
    return -1;
/*
 * Copy the number into the work1 array.
 */
  if(ccb_ull_copy(a, na, work1, nwork1))
    return -1;
/*
 * Since we don't know how long the integer is going to be in advance, but
 * we need to start with the rightmost digit, start to compose the string
 * from the last character available in the output buffer. Later we will
 * shift its least significant bit to the least significant bit of the
 * answer.
 */
  s[dim-1] = '\0';
  while(!ccb_ull_is_zero(work1, nwork1)) {
    int c;   /* The latest digit character to be output */
/*
 * Would adding another character to the output string overflow
 * the bounds of the string?
 */
    if(nchar+1 >= dim)
      return -1;
/*
 * Divide the remaining part of the number by 10, and get the remainder.
 */
    if(ccb_ull_divide(work1, nwork1, &ten, 1, work2, nwork2, work1, nwork1))
      return -1;
/*
 * Convert the remainder to the corresponding digit character.
 */
    c = digit_of_value(work1[0]);
/*
 * Prepend the new digit character to the previous characters of the
 * integer in the output string.
 */
    s[dim - (++nchar + 1)] = c;
/*
 * Copy the result of the division into work1[]. This contains the
 * remaining digits of the integer still to be written.
 */
    if(ccb_ull_copy(work2, nwork2, work1, nwork1))
      return -1;
  };
/*
 * If no characters were written to the string, this means that
 * the number was zero.
 */
  if(nchar < 1)
    s[dim - (++nchar + 1)] = '0';
/*
 * Move the start of the number to the start of the string.
 */
  memmove(s, s+dim-(nchar+1), nchar+1);
  return nchar;
}

/*.......................................................................
 * Convert a value between 0 and 9, to the numeric digit that represents
 * it in text.
 *
 * Input:
 *  value   unsigned  The value to be encoded.
 * Output:
 *  return       int  The digit character that represents the value,
 *                    or '\0' if the value isn't in the range 0..9.
 */
static int digit_of_value(int value)
{
/*
 * Lookup the digit character that represents 'value'.
 */
  switch(value) {
  case 0:
    return '0';
  case 1:
    return '1';
  case 2:
    return '2';
  case 3:
    return '3';
  case 4:
    return '4';
  case 5:
    return '5';
  case 6:
    return '6';
  case 7:
    return '7';
  case 8:
    return '8';
  case 9:
    return '9';
  };
/*
 * Not a digit.
 */
  return '\0';
}
