#ifndef input_h
#define input_h

#include <stdarg.h>
#include <stdio.h>  /* (FILE *) */

/* Declare the tag for a generic input stream. */

typedef struct InputStream InputStream;

/* List iterator-specific method function declarations */

/*
 * Read a single character from an input source and return it.
 *
 * Input:
 *  data          void *  The context data of the input source.
 * Output:
 *  return         int    The newly read character, or EOF on
 *                        error.
 */
#define INPUT_READ_FN(fn) int (fn)(void *data)

/*
 * Delete the context data of an input source.
 *
 * Input:
 *  data          void *  The context data to be deleted.
 * Output:
 *  return        void *  The deleted context data (always NULL).
 */
#define INPUT_DEL_FN(fn) void *(fn)(void *data)

/*
 * Perform the equivalent of clearerr() on a source.
 *
 * Input:
 *  data            void *  The context data of the input
 *                          source to be reset.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Couldn't clear error status.
 */
#define INPUT_CLR_FN(fn) int (fn)(void *data)

/*
 * Report the source and position of the stream to stderr in a form
 * that is suitable to be used as the prefix of an error message.
 *
 * Input:
 *  stream  InputStream *  The parent input stream.
 *  data           void *  The context data of the input source
 *                         that suffered the error.
 *  errmsg         char *  The buffer in which to write.
 *  errdim       size_t    The dimension of errmsg[].
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error.
 */
#define INPUT_ERR_FN(fn) int (fn)(InputStream *stream, void *data, \
				  char *errmsg, size_t errdim)

/* Set the maximum length of any input token */

enum {INPUT_WORKLEN=1024};

/* Set the length of the error-message buffer */

enum {INPUT_ERR_LEN=128};

/*
 * Construct an initially closed input-stream, ready to be connected
 * to an input source with open_InputStream().
 */
InputStream *new_InputStream(void);

/*
 * Declare a generic stream destructor.
 * Note that this will call close_InputStream() if the stream is
 * connected to an input source.
 */
InputStream *del_InputStream(InputStream *stream);

/*
 * Connect an input stream to a specific input source. This function
 * should only be called from type-specific stream open_*()
 * functions (see below). Note that if another input source is currently
 * connected, it will be pushed down a stack of input sources, and reading
 * will resume from it when the end of new input source has been reached.
 */
int open_InputStream(InputStream *stream, void *data,
		     INPUT_READ_FN(*read_fn), INPUT_CLR_FN(*clr_fn),
		     INPUT_ERR_FN(*err_fn), INPUT_DEL_FN(*del_fn));

/*
 * Close all input sources of an input stream. Future reads will return
 * an error until the stream is successfully re-opened to another input source.
 */
void close_InputStream(InputStream *stream);

/*
 * Declare functions that connect generic input streams to
 * specific types of input sources.
 */
int open_FileInputStream(InputStream *stream, const char *dir,
			 const char *name);
int open_StringInputStream(InputStream *stream, int copy, const char *string);
int open_StdioInputStream(InputStream *stream, int do_close, FILE *fp);

/*
 * The following function resets a stream by clearing its error
 * flags if possible, and reading up to the first character of
 * the next line. On failure it returns non-zero.
 */
int reset_InputStream(InputStream *stream);

/*
 * The following function reads the next character from the input stream.
 * In the process it removes escaped newline characters and comments.
 */
int read_InputStream(InputStream *stream);

/*
 * Peek at the next character in the stream, without removing it from
 * the stream.
 */
int input_peek_char(InputStream *stream);

/*
 * Get the last error message that was recorded by the stream.
 */

/*
 * Canned lexical input functions.
 */
int input_skip_to_eol(InputStream *stream);
int input_skip_past_eol(InputStream *stream);
int input_skip_space(InputStream *stream, int advance);
int input_skip_white(InputStream *stream, int advance);
int input_skip_char(InputStream *stream);

/*
 * Read an ASCII keyword from an input stream. Keywords must commence
 * with an alphabetical character continue with alphanumeric
 * and underscore characters. The output keyword will be left in
 * stream->work[], where it will have been converted to lower case
 * if the fold argument is true.
 */
int input_keyword(InputStream *stream, int fold);

/*
 * Read a string that is enclosed in double quotes. All C-style
 * escape sequences (eg. \n) are recognized. The string will be
 * left in stream->work[].
 */
int input_quoted_string(InputStream *stream);

/*
 * The following macro defines the prototype of functions like
 * isspace(), to be used by input_literal() to identify characters
 * to include in the string.
 */
#define IS_LITERAL_FN(fn) int (fn)(int c)

/*
 * Read an unquoted string. The chararacters listed in opn[] and cls[]
 * are treated as open and close parentheses. The string will be
 * terminated when the provided is_literal() function returns 0
 * outside of sub-strings and unmatched parentheses. The only
 * exception to this is if the terminating character is a newline
 * character and this character was preceded (ignoring intervening
 * spaces) by one of the characters in nl_escapes[]. The string will
 * be left in stream->work[]. The terminating character will be left
 * in the stream.
 */
int input_literal(InputStream *stream, char *opn, char *cls,
		  IS_LITERAL_FN(*is_literal), char *nl_escapes);

/*
 * Read an ASCII word from an input stream. A word is defined as a string
 * of printable characters excluding white-space. If fold is true, fold
 * upper case characters to lower case. The string will be left in
 * stream->work[].
 */
int input_word(InputStream *stream, int fold);

/*
 * Read a signed integer. If anybase is true, interpret a number
 * that starts with a zero as octal, one that starts with 0x as
 * hexadecimal and one that starts with 0b as binary. Otherwise only
 * accept base 10 numbers.
 */
int input_long(InputStream *stream, int anybase, long *lval);

/*
 * Read an unsigned integer. If anybase is true, interpret a number
 * that starts with a zero as octal, one that starts with 0x as
 * hexadecimal and one that starts with 0b as binary. Otherwise only
 * accept base 10 numbers.
 */
int input_ulong(InputStream *stream, int anybase, unsigned long *ulval);

/*
 * Read a double precision number.
 */
int input_double(InputStream *stream, double *dval);

typedef struct {
  enum {
    NUM_DOUBLE,   /* Number was written as an int with no exponent */
    NUM_INT       /* Number was written as floating point and/or with exponent*/
  } type;
  int sign;       /* Sign of number as -1 or 1 */
  union {
    double dval;  /* Use if type==NUM_DOUBLE */
    int ival;     /* Use if type==NUM_INT */
  } value;
} Number;

int input_number(InputStream *stream, int sign_ok, Number *number);

/*
 * Read one or more sexagesimal components separated by ':' characters.
 * The first number can have any integral value and can be signed. The
 * last number can have a fractional part. The intervening parts must
 * be integers between 0 and 59. The resulting number has the dimensions
 * of the first number. Thus 23:30 can also be entered as 23.5.
 */
int input_sexagesimal(InputStream *stream, double *result);

/*
 * Read a date like 23-JAN-1997.
 */
int input_date(InputStream *stream, int *year, int *month, int *day);

/*
 * Read a time from an input stream. The following are valid time
 * specifications along with their interpretations:
 *
 *  23:34:04.5  ->   23:34:04.5
 *  23:34:4.5   ->   23:34:04.5
 *  23:34:10    ->   23:34:10
 *  23:34       ->   23:34:00
 *  23          ->   23:00:00
 */
int input_time(InputStream *stream, int *hour, int *min, double *sec);

/*
 * Read a date and optional time. The following are valid specifications
 * and their interpretations.
 *
 *  12-DEC-1998                 ->  12-DEC-1998 00:00:00.0
 *  12-DEC-1998:13              ->  12-DEC-1998 13:00:00.0
 *  12-DEC-1998:13:34           ->  12-DEC-1998 13:34:00.0
 *  12-DEC-1998:13:34:23        ->  12-DEC-1998 13:34:23.0
 *  12-DEC-1998:13:34:23.5      ->  12-DEC-1998 13:34:23.5
 *
 * The following are also valid if nospace!=0.
 *
 *  12-DEC-1998 13              ->  12-DEC-1998 13:00:00.0
 *  12-DEC-1998 13:34           ->  12-DEC-1998 13:34:00.0
 *  12-DEC-1998 13:34:23        ->  12-DEC-1998 13:34:23.0
 *  12-DEC-1998 13:34:23.5      ->  12-DEC-1998 13:34:23.5
 */
int input_date_and_time(InputStream *stream, int nospace, int *year,
			int *month, int *day, int *hour, int *min, double *sec);

/*
 * Read a time interval written like the following example:
 *
 *  23d:14h:16m:2.32s   (ie. 23 days, 14 hours, 16 minutes, 2.32 seconds)
 *
 * Any of the components can be omitted, but each trailing colon is taken
 * as evidence that another component follows. The result is returned in
 * seconds, via the *interval argument.
 */
int input_interval(InputStream *stream, double *interval);

/*-----------------------------------------------------------------------
 * Error management conventions and facilities.
 *
 * When the above input_whatever() functions detect an error they
 * place an appropriate error message in the internal error reporting
 * buffer of the input stream, then return non-zero to indicate that
 * an error occured. The input_error() function is designed with this
 * in mind.  It is a printf-style function that returns 1 for use as a
 * non-zero error code. One statement can then be used both to record
 * the error and to return from the calling function with an error
 * code. For example:
 *
 * int input_ulong(InputStream *stream, unsigned long *ul)
 * {
 *   if(!isdigit(input_peek_char(stream)))
 *    return input_error(stream, "Not an unsigned integer");
 *   ...
 *  }
 *
 * Where possible input_error() reports the location of the errant input.
 * For example, if the input stream is connected to a file then it reports
 * the file name and line number at which the error was encountered.
 *
 * The error message associated with the last error, can subsequently be
 * retrieved by calling last_input_error(stream).
 */
int input_error(InputStream *stream, const char *fmt, ...)
#ifdef __GNUC__
__attribute__ ((format (printf, 2, 3))); /* Get gcc to check the format etc.. */
#else
     ;
#endif

/*
 * Provide a version of input_error() that takes an already encapsulated
 * variable argument list.
 */
int input_verror(InputStream *stream, const char *fmt, va_list args);

/*
 * Retrieve a copy of the message associated with the last error input
 * error. If a buffer is provided, the contents of the internal
 * error-message buffer are copied into it, and the return value is a
 * pointer to the caller's buffer. Otherwise a pointer to the internal
 * error message buffer of the stream is returned. The latter should
 * be assumed to be invalid on the next call to any function that
 * operates on the specified stream.
 */
const char *last_input_error(InputStream *stream, char *errmsg, size_t errdim);

/*
 * Retrieve a copy of the last token read. If a buffer is provided,
 * the contents of the internal token buffer are copied into it, and
 * the return value is a pointer to the caller's buffer. Otherwise a
 * pointer to the internal token buffer of the stream is returned. The
 * latter should be assumed to be invalid on the next call to any
 * function that operates on the specified stream.
 */
const char *last_input_token(InputStream *stream, char *buffer, size_t bufdim);


#endif
